#!/usr/bin/env python3
"""
a common class for pegasus workflows that work on VCF variant files
"""
import sys, os
import copy
import logging
from typing import Dict, List
from pegaflow.api import File, Transformation
from palos import Genome, PassingData, utils
from palos.io.MatrixFile import MatrixFile
from palos.ngs.io.VCFFile import VCFFile
from palos import ngs
from . MapReduceGenomeFileWorkflow import MapReduceGenomeFileWorkflow
ParentClass = MapReduceGenomeFileWorkflow

class AbstractVCFWorkflow(ParentClass):
    __doc__ = __doc__
    option_default_dict = copy.deepcopy(ParentClass.option_default_dict)
    option_default_dict.update({
        ('minDepth', 0, float): [0, 'm', 1,
            'minimum depth for a call to regarded as non-missing', ],\
        ("ligateVcfPerlPath", 1, ): ["%s/bin/umake/scripts/ligateVcf.pl", '', 1, 
            'path to ligateVcf.pl'],\
        ("notToKnowNoOfLoci", 0, int): [0, '', 0, 
            'By default, this program infers the number of loci for each VCF '
            'file (for splitting, etc.). '
            'Either use the first number of filename as db id or read through '
            'the file. Toggle this to disable it.'],\
        ("notToUseDBToInferVCFNoOfLoci", 0, int): [0, '', 0, 
            'Assuming --notToKnowNoOfLoci is off, '
            'toggle this to infer the number of loci by strictly reading through the file'],\
        ('minNoOfLociInVCF', 0, int): [5, '', 1, 
            'minimum number of loci for an input VCF file to be included.', ],\
        })
    #update these two as they mean the number of loci now, not base pairs
    option_default_dict[('intervalOverlapSize', 1, int)][0] = 100
    option_default_dict[('intervalSize', 1, int)][0] = 10000
    option_default_dict[('inputSuffixList', 0, )][0] = ".vcf"
    #default input suffix is .vcf
    
    def __init__(self,  **keywords):
        """
        """
        ParentClass.__init__(self, **keywords)
        if getattr(self, "input_path", None):
            self.input_path = os.path.abspath(self.input_path)
        
        if hasattr(self, 'ligateVcfPerlPath'):
            self.ligateVcfPerlPath = self.insertHomePath(
                self.ligateVcfPerlPath, self.home_path)

    def addAddVCFFile2DBJob(self, executable=None, inputFile=None,
        genotypeMethodShortName=None, logFile=None, format=None,
        data_dir=None, checkEmptyVCFByReading=None, commit=False,
        parentJobLs=None, extraDependentInputLs=None, transferOutput=False,
        extraArguments=None, job_max_memory=2000, **keywords):
        """
        """
        extraArgumentList = ['--format', format]
        if logFile:
            extraArgumentList.extend(["--logFilename", logFile])
        if data_dir:
            extraArgumentList.extend(['--data_dir', data_dir])
        if checkEmptyVCFByReading:
            extraArgumentList.extend(['--checkEmptyVCFByReading'])
        if genotypeMethodShortName:
            extraArgumentList.extend(['--genotypeMethodShortName', \
                genotypeMethodShortName, ])
        if commit:
            extraArgumentList.append('--commit')
        if extraArguments:
            extraArgumentList.append(extraArguments)
        
        job= self.addDBJob(executable=executable, inputFile=inputFile,
            outputFile=None,
            parentJobLs=parentJobLs,
            extraDependentInputLs=extraDependentInputLs,
            extraOutputLs=[logFile],
            transferOutput=transferOutput, \
            extraArgumentList=extraArgumentList, job_max_memory=job_max_memory,
            **keywords)
        return job
    
    def registerExecutables(self):
        """
        """
        ParentClass.registerExecutables(self)
        
        #2012.8.30 moved from vervet/src/AddVCFFolder2DBWorkflow.py
        #self.registerOneExecutable(path=os.path.join(self.pymodulePath, \
        #	"db/import/AddVCFFile2DB.py"),
        #	name='AddVCFFile2DB', clusterSizeMultiplier=1)
        
        self.registerOneExecutable(path=os.path.join(self.pymodulePath, \
            "mapper/filter/FilterVCFSNPCluster.py"),
            name='FilterVCFSNPCluster', clusterSizeMultiplier=1)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath, \
            "mapper/extractor/JuxtaposeAlleleFrequencyFromMultiVCFInput.py"),
            name='JuxtaposeAlleleFrequencyFromMultiVCFInput', clusterSizeMultiplier=1)		
        
        #2013.07.12
        self.registerOneExecutable(path=self.javaPath, name='SelectVariantsJavaInReduce', \
            clusterSizeMultiplier=0.001)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath, 
            "polymorphism/qc/RemoveRedundantLociFromVCF.py"), \
            name='RemoveRedundantLociFromVCF_InReduce', clusterSizeMultiplier=0)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath, 
            "polymorphism/qc/RemoveRedundantLociFromVCF.py"), \
            name='RemoveRedundantLociFromVCF', clusterSizeMultiplier=1)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath, 
            "polymorphism/qc/ClearVCFBasedOnSwitchDensity.py"), \
            name='ClearVCFBasedOnSwitchDensity', clusterSizeMultiplier=1)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath, 
            "polymorphism/qc/CalculateSameSiteConcordanceInVCF.py"), \
            name='CalculateSameSiteConcordanceInVCF', clusterSizeMultiplier=1)
        
        self.registerOneExecutable(path=os.path.join(self.pymodulePath, 
            "mapper/extractor/ExtractInfoFromVCF.py"), \
            name='ExtractInfoFromVCF', clusterSizeMultiplier=1)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath, 
            "mapper/extractor/ExtractSamplesFromVCF.py"), \
            name='ExtractSamplesFromVCF', clusterSizeMultiplier=1)
    
    def registerCommonExecutables(self):
        """
        """
        ParentClass.registerCommonExecutables(self)
    
    def registerFilesOfInputDir(self, inputDir=None, input_site_handler=None, \
        checkEmptyVCFByReading=False, pegasusFolderName='',\
        maxContigID=None, minContigID=None, db_main=None, needToKnowNoOfLoci=False,
        minNoOfLociInVCF=None, includeIndelVCF=True, notToUseDBToInferVCFNoOfLoci=None):
        """
        2013.3.1 flip includeIndelVCF to true.
            now indel and SNP vcf files from AlignmentToCall workflows are in separate folders.
        Argument db_main, needToKnowNoOfLoci, to get noOfLoci by parsing inputFname and find db-entry...
            argument minNoOfLociInVCF, only used when it's not None and needToKnowNoOfLoci is True
        Register the tbi file if it exists.
        The returning data structure is changed to conform to some standard used across several functions.
        vcf files only.
        """
        if input_site_handler is None:
            input_site_handler = self.input_site_handler
        if notToUseDBToInferVCFNoOfLoci:
            notToUseDBToInferVCFNoOfLoci = self.notToUseDBToInferVCFNoOfLoci
        
        sys.stderr.write("Registering input files from %s, maxContigID=%s, minContigID=%s, "
            "needToKnowNoOfLoci=%s, minNoOfLociInVCF=%s, includeIndelVCF=%s, notToUseDBToInferVCFNoOfLoci=%s ..."%\
            (inputDir, maxContigID, minContigID, needToKnowNoOfLoci,
            minNoOfLociInVCF, includeIndelVCF, notToUseDBToInferVCFNoOfLoci))
        returnData = PassingData(jobDataLs = [])
        counter = 0
        real_counter = 0
        if inputDir and os.path.isdir(inputDir):
            fnameLs = os.listdir(inputDir)
            previous_reported_real_counter = ''
            for fname in fnameLs:
                counter += 1
                inputFname = os.path.realpath(os.path.join(inputDir, fname))
                if (maxContigID is not None and maxContigID!=0) or (minContigID is not None and minContigID!=0):
                    try:
                        contigID = int(self.getContigIDFromFname(os.path.basename(fname)))
                        if (maxContigID is not None and maxContigID!=0) and contigID>maxContigID:
                            continue
                        if (minContigID is not None and minContigID!=0) and contigID<minContigID:
                            continue
                    except:
                        sys.stderr.write('Except type in handling file %s: %s\n'%(inputFname, repr(sys.exc_info())))
                        import traceback
                        traceback.print_exc()
                        # Ignore VCFs that contig/chromosome IDs could not be derived. 
                        continue
                if ngs.isFileNameVCF(fname, includeIndelVCF=includeIndelVCF) and \
                        not ngs.isVCFFileEmpty(inputFname, checkContent=checkEmptyVCFByReading):
                    no_of_loci = None
                    no_of_individuals = None
                    if needToKnowNoOfLoci:
                        if notToUseDBToInferVCFNoOfLoci:
                            # Do not use db to infer no of loci
                            pass
                        elif db_main:
                            genotype_file = db_main.parseGenotypeFileGivenDBAffiliatedFilename(filename=inputFname)
                            if genotype_file and inputFname.find(genotype_file.path)>=0:
                                # make sure same file
                                no_of_loci = genotype_file.no_of_loci
                                no_of_individuals = genotype_file.no_of_individuals
                        if no_of_loci is None:
                            #do file parsing
                            vcfFile = VCFFile(inputFname=inputFname, report=False)
                            no_of_loci = vcfFile.getNoOfLoci()
                            no_of_individuals = len(vcfFile.getSampleIDList())
                            vcfFile.close()
                    
                    if minNoOfLociInVCF is None or no_of_loci is None or \
                            (minNoOfLociInVCF is not None and no_of_loci is not None and \
                                no_of_loci >=minNoOfLociInVCF):
                        input_file = self.registerOneInputFile(input_path=inputFname, folderName=pegasusFolderName)
                        inputBaseFname = os.path.basename(inputFname)
                        input_file.noOfLoci = no_of_loci
                        input_file.no_of_loci = no_of_loci
                        input_file.no_of_individuals = no_of_individuals
                        input_file.noOfIndividuals = no_of_individuals
                        tbi_F_absPath = "%s.tbi"%inputFname
                        if os.path.isfile(tbi_F_absPath):	#it exists
                            tbi_F = self.registerOneInputFile(input_path=tbi_F_absPath,
                                folderName=pegasusFolderName)
                        else:
                            tbi_F = None
                        input_file.tbi_F = tbi_F
                        returnData.jobDataLs.append(PassingData(job=None, jobLs=[], \
                            vcfFile=input_file, tbi_F=tbi_F, file=input_file, fileLs=[input_file, tbi_F]))
                        real_counter += 1
                    if real_counter%10==0:
                        sys.stderr.write("%s%s"%('\x08'*len(previous_reported_real_counter), real_counter))
                        previous_reported_real_counter = repr(real_counter)
        sys.stderr.write("  %s non-empty VCF out of %s files, real_counter=%s.\n"%(
            len(returnData.jobDataLs),
            counter, real_counter))
        return returnData
    
    def addSplitVCFFileJob(self, executable:Transformation=None,
        inputFile:File=None, outputFnamePrefix=None,
        noOfOverlappingSites=1000, noOfSitesPerUnit=5000, noOfTotalSites=10000,
        extraArguments=None,
        parentJobLs=None,
        extraDependentInputLs=None,
        transferOutput=True, job_max_memory=2000,
        sshDBTunnel=False, **keywords):
        """
        """
        if extraDependentInputLs is None:
            extraDependentInputLs = []
        
        #turn them into nonnegative	
        noOfSitesPerUnit = abs(noOfSitesPerUnit)
        noOfOverlappingSites = abs(noOfOverlappingSites)
        if noOfTotalSites is not None:
            noOfTotalSites = abs(noOfTotalSites)
        else:
            noOfTotalSites = 10000000	#make it really big ,so no split
            noOfSitesPerUnit = noOfTotalSites
        
        if noOfSitesPerUnit>noOfTotalSites:
            noOfSitesPerUnit = noOfTotalSites
        if noOfOverlappingSites>noOfSitesPerUnit:
            noOfOverlappingSites = noOfSitesPerUnit
        key2ObjectForJob = {}
        extraArgumentList = ["-O %s"%outputFnamePrefix,]
        extraOutputLs = []
        if noOfOverlappingSites is not None:
            extraArgumentList.append('--noOfOverlappingSites %s'%(noOfOverlappingSites))
        if noOfSitesPerUnit is not None:
            extraArgumentList.append('--noOfSitesPerUnit %s'%(noOfSitesPerUnit))
        if noOfTotalSites is not None:
            extraArgumentList.append('--noOfTotalSites %s'%(noOfTotalSites))			
        noOfUnits = max(1, utils.getNoOfUnitsNeededToCoverN(N=noOfTotalSites,
            s=noOfSitesPerUnit, o=noOfOverlappingSites)-1)
        
        suffixAndNameTupleList = []
        # a list of tuples , in each tuple, 1st element is the suffix. 2nd element is the proper name of the suffix.
        #job.$nameFile will be the way to access the file.
        #if 2nd element (name) is missing, suffix[1:].replace('.', '_') is the name (dot replaced by _) 
        for i in range(1, noOfUnits+1):
            suffixAndNameTupleList.append(['_unit%s.vcf'%(i), 'unit%s'%(i)])
        if extraArguments:
            extraArgumentList.append(extraArguments)
        self.setupMoreOutputAccordingToSuffixAndNameTupleList(outputFnamePrefix=outputFnamePrefix, \
            suffixAndNameTupleList=suffixAndNameTupleList, \
            extraOutputLs=extraOutputLs, key2ObjectForJob=key2ObjectForJob)
        
        job = self.addGenericJob(executable=executable, inputFile=inputFile, outputFile=None, \
            extraArgumentList=extraArgumentList,
            parentJobLs=parentJobLs, extraDependentInputLs=extraDependentInputLs, \
            extraOutputLs=extraOutputLs,\
            transferOutput=transferOutput, \
            key2ObjectForJob=key2ObjectForJob, job_max_memory=job_max_memory, \
            sshDBTunnel=sshDBTunnel, **keywords)
        job.noOfSitesPerUnit = noOfSitesPerUnit
        job.noOfUnits  = noOfUnits
        return job
    
    
    def addLigateVcfJob(self, executable:Transformation=None,
        ligateVcfExecutableFile:File=None, outputFile:File=None, \
        extraArguments=None,
        parentJobLs=None, extraDependentInputLs=None, transferOutput=False, \
        job_max_memory=2000, **keywords):
        """
        2013.07.04 moved from AlignmentToCallPipeline.py
        2012.06.26
            rename argument ligateVcfPerlPath to ligateVcfExecutableFile
        2012.6.1
            ligateVcf ligates overlapping VCF files.
        """
        if extraDependentInputLs is None:
            extraDependentInputLs = []
        extraArgumentList = [ligateVcfExecutableFile, outputFile]
        extraDependentInputLs.append(ligateVcfExecutableFile)
        if extraArguments:
            extraArgumentList.append(extraArguments)
        #do not pass outputFile as argument to addGenericJob() because addGenericJob will add "-o" in front of it.
        return self.addGenericJob(executable=executable, \
            extraArgumentList=extraArgumentList,
            extraDependentInputLs=extraDependentInputLs,
            extraOutputLs=[outputFile], \
            parentJobLs=parentJobLs,
            transferOutput=transferOutput, \
            job_max_memory=job_max_memory, **keywords)
        
    
    def getChr2IntervalDataLsBySelectVCFFile(self, vcfFname=None,
        noOfSitesPerUnit=5000, noOfOverlappingSites=1000, \
        folderName=None, parentJobLs= None):
        """
        2012.8.9 update it so that the interval encompassing all lines in one block/unit is known.
            good for mpileup to only work on that interval and then "bcftools view" select from sites from the block.
            TODO: offer partitioning by equal-chromosome span, rather than number of sites.
            Some sites could be in far from each other in one block,
                which could incur long-running mpileup. goal is to skip these deserts.
        Bugfix: add -1 to the starting number below cuz otherwise it's included in the next block's start
            blockStopLineNumber = min(startLineNumber+(i+1)*noOfSitesPerUnit-1, stopLineNumber)	
        2012.8.14
            1.
            2. folderName is the relative path of the folder in the pegasus workflow, that holds vcfFname.
                it'll be created upon file stage-in. no mkdir job for it.
                
            get the number of lines in vcfFname.
            get chr2StartStopDataLsTuple
            for each chromosome, split its lines into units  that don't exceed noOfSitesPerUnit
                add the split job
                 
        """
        sys.stderr.write("Splitting %s into blocks of %s lines ... "%(vcfFname, noOfSitesPerUnit))
        #from palos import utils
        #noOfLines = utils.getNoOfLinesInOneFileByWC(vcfFname)
        chr2StartStopDataLs = {}
        reader = MatrixFile(vcfFname)
        #csv.reader(inf, delimiter=figureOutDelimiter(vcfFname))
        lineNumber = 0
        previousChromosome = None
        previousLine = None
        chromosome = None
        for row in reader:
            lineNumber += 1
            chromosome, start, stop = row[:3]
            start = int(start)	#0-based, starting base
            stop = int(stop)	#0-based, stopping base but not inclusive, i.e. [start, stop)
            
            if previousLine is None or chromosome!=previousLine.chromosome:
                #first line or different chromosome
                if previousLine is not None and previousLine.chromosome is not None:
                    
                    prevChrLastStartStopData = chr2StartStopDataLs[previousLine.chromosome][-1]
                    if prevChrLastStartStopData.stopLineNumber is None:
                        prevChrLastStartStopData.stopLineNumber = previousLine.lineNumber
                        prevChrLastStartStopData.stopLineStart = previousLine.start
                        prevChrLastStartStopData.stopLineStop = previousLine.stop
                    
                if chromosome not in chr2StartStopDataLs:
                    StartStopData = PassingData(startLineNumber=lineNumber,
                        startLineStart=start, startLineStop=stop,
                        stopLineNumber=None, stopLineStart=None, stopLineStop=None)
                    chr2StartStopDataLs[chromosome] = [StartStopData]
            else:	#same chromosome and not first line
                lastStartStopData = chr2StartStopDataLs[chromosome][-1]
                if lastStartStopData.stopLineNumber is None:
                    #last block hasn't been closed yet.
                    noOfLinesInCurrentBlock = lineNumber - lastStartStopData.startLineNumber +1
                    if noOfLinesInCurrentBlock>=noOfSitesPerUnit:
                        #time to close it
                        lastStartStopData.stopLineNumber = lineNumber
                        lastStartStopData.stopLineStart = start
                        lastStartStopData.stopLineStop = stop
                else:	#generate a new block
                    StartStopData = PassingData(startLineNumber=lineNumber,
                        startLineStart=start, startLineStop=stop,
                        stopLineNumber=None, stopLineStart=None,
                        stopLineStop=None)
                    chr2StartStopDataLs[chromosome].append(StartStopData)
            previousLine = PassingData(chromosome = chromosome, start=start,
                stop=stop, lineNumber=lineNumber)
        #final closure
        if previousLine is not None:	#vcfFname is not empty
            lastStartStopData = chr2StartStopDataLs[previousLine.chromosome][-1]
            if lastStartStopData.stopLineNumber is None:	#last block hasn't been closed yet.
                #close it regardless of whether it has enough lines in it or not.
                lastStartStopData.stopLineNumber = previousLine.lineNumber
                lastStartStopData.stopLineStart = previousLine.start
                lastStartStopData.stopLineStop = previousLine.stop
        sys.stderr.write("%s chromosomes out of %s lines.\n"%(len(chr2StartStopDataLs), lineNumber))
        
        intervalFile = self.registerOneInputFile(vcfFname, folderName=folderName)
        chr2IntervalDataLs = {}
        counter = 0
        for chromosome, startStopDataLs in chr2StartStopDataLs.items():
            for startStopData in startStopDataLs:
                blockStartLineNumber = startStopData.startLineNumber
                blockStopLineNumber = startStopData.stopLineNumber
                # 2012.8.9 the large interval that encompasses all BED lines 
                interval = '%s:%s-%s'%(chromosome, startStopData.startLineStart, startStopData.stopLineStop)
                blockIntervalFile = File(os.path.join(folderName,
                    '%s_line_%s_%s_bed.tsv'%(chromosome,
                    blockStartLineNumber, blockStopLineNumber)))
                blockIntervalJob = self.addSelectLineBlockFromFileJob(
                    executable=self.SelectLineBlockFromFile,
                    inputFile=intervalFile, outputFile=blockIntervalFile,
                    startLineNumber=blockStartLineNumber,
                    stopLineNumber=blockStopLineNumber,
                    parentJobLs=parentJobLs, extraDependentInputLs=None,
                    transferOutput=False, job_max_memory=500)
                intervalFileBasenameSignature = '%s_%s_%s'%(
                    chromosome, blockStartLineNumber, blockStopLineNumber)
                if chromosome not in chr2IntervalDataLs:
                    chr2IntervalDataLs[chromosome] = []
                intervalData = PassingData(file=blockIntervalFile,
                    intervalFileBasenameSignature=intervalFileBasenameSignature,
                    interval=interval,
                    chr=chromosome, chromosome=chromosome, jobLs=[blockIntervalJob],
                    job=blockIntervalJob)
                chr2IntervalDataLs[chromosome].append(intervalData)
                counter += 1
        sys.stderr.write("%s intervals and %s SelectLineBlockFromFile jobs.\n"%(counter, counter))
        return chr2IntervalDataLs
    
    
    def mapEachInterval(self, VCFJobData=None, chromosome=None, intervalData=None,
        mapEachChromosomeData=None, passingData=None, transferOutput=False,
        **keywords):
        """
        argument VCFJobData looks like PassingData(file=splitVCFFile,
            vcfFile=splitVCFFile, fileLs=[splitVCFFile], \
            job=splitVCFJob, jobLs=[splitVCFJob], tbi_F=None)
        """
        
        returnData = PassingData(no_of_jobs = 0)
        returnData.jobDataLs = []
        passingData.intervalFileBasenamePrefix
        passingData.splitVCFFile
        passingData.unitNumber
        """
        ## 2013.06.19 structures available from passingData, specific to the interval
        passingData.splitVCFFile = splitVCFFile
        passingData.unitNumber = unitNumber
        passingData.intervalFileBasenamePrefix = '%s_%s_splitVCF_u%s'%(chromosome, commonPrefix, unitNumber)
        passingData.noOfIndividuals = jobData.file.noOfIndividuals
        passingData.span = self.intervalSize + self.intervalOverlapSize*2 
        #2013.06.19 for memory/walltime gauging
        """
        return returnData
    
    def mapEachVCF(self, chromosome=None,VCFJobData=None, passingData=None,
        transferOutput=False, **keywords):
        """
        new data structure available used to split/select VCF
            intervalDataLs = passingData.intervalDataLs
        default is to split each VCF into intervals
        """
        
        returnData = PassingData(no_of_jobs = 0)
        returnData.jobDataLs = []
        
        topOutputDirJob = passingData.topOutputDirJob
        fileBasenamePrefix = passingData.fileBasenamePrefix
        VCFJobData = passingData.VCFJobData
        VCFFile = VCFJobData.file	#2013.04.08
        jobData = passingData.jobData
        intervalOverlapSize = passingData.intervalOverlapSize
        intervalSize = passingData.intervalSize
        
        #intervalDataLs = passingData.intervalDataLs
        # #2013.06.20 new data structure available used to split/select VCF
        outputFnamePrefix = os.path.join(topOutputDirJob.output, '%s_splitVCF'%fileBasenamePrefix)
        
        splitVCFJob = self.addSplitVCFFileJob(
            executable=self.SplitVCFFile,
            inputFile=VCFFile,
            outputFnamePrefix=outputFnamePrefix,
            noOfOverlappingSites=intervalOverlapSize,
            noOfSitesPerUnit=intervalSize,
            noOfTotalSites=getattr(VCFFile, 'noOfLoci', None),
            parentJobLs=jobData.jobLs+[topOutputDirJob],
            extraDependentInputLs=[jobData.tbi_F],
            extraArguments=None, transferOutput=transferOutput,
            job_max_memory=2000)
        self.no_of_jobs +=1
        returnData.jobDataLs.append(PassingData(jobLs=[splitVCFJob],
            file=splitVCFJob.output,
            fileLs=splitVCFJob.outputLs))
        returnData.splitVCFJob = splitVCFJob
        
        return returnData
    
    def hierarchicalCombineIntervalJobIntoOneChromosome(self, unionJob=None,
        refFastaFList=None, fileBasenamePrefix=None,
        passingData=None, intervalJobLs=None,
        maxNoOfIntervalJobsInOneUnion=500,
        outputDirJob=None,
        needBGzipAndTabixJob=False, 
        transferOutput=False, job_max_memory=None, walltime=None,
        **keywords):
        """
        #. Given a unionJob, a two-step VCF concatenation function
            #. 1st-step, concatenate a continuous string of intervals into big-intervals
            #. 2nd-step, add the big-interval jobs into the final union job.
        e.g.
        #combine intervals hierarchically, to avoid open file number limit on cluster
        self.hierarchicalCombineIntervalJobIntoOneChromosome(
            unionJob=combineChromosomeJobData.callUnionJob, \
            refFastaFList=refFastaFList, fileBasenamePrefix="subUnion",\
            intervalJobLs=chromosome2jobData[chromosome].genotypingJobLs, \
            maxNoOfIntervalJobsInOneUnion=250,\
            outputDirJob=combineChromosomeJobData.callOutputDirJob,
            transferOutput=False, job_max_memory=None, walltime=120,
            needBGzipAndTabixJob=False)
        This function goes around the limitation in the number of open files on cluster.
            combine a long list of interval jobs in 2 steps.
        """
        for i in range(0, len(intervalJobLs), maxNoOfIntervalJobsInOneUnion):
            subIntervalJobLs = intervalJobLs[i:i+maxNoOfIntervalJobsInOneUnion]
            firstIntervalJob = subIntervalJobLs[0]
            lastIntervalJob = subIntervalJobLs[-1]
            chromosome = firstIntervalJob.intervalData.chromosome
            startPos = firstIntervalJob.intervalData.start
            stopPos = lastIntervalJob.intervalData.stop
            subUnionJobData = self.concatenateIntervalsIntoOneVCFSubWorkflow(
                refFastaFList=refFastaFList, \
                fileBasenamePrefix="%s_%s_%s_%s"%(fileBasenamePrefix, chromosome, startPos, stopPos),
                passingData=passingData, intervalJobLs=subIntervalJobLs,
                outputDirJob=outputDirJob,
                needBGzipAndTabixJob=needBGzipAndTabixJob,
                transferOutput=transferOutput, job_max_memory=job_max_memory,
                walltime=walltime,
                **keywords)
            self.addInputToMergeJob(unionJob, inputF=subUnionJobData.file,
                parentJobLs=subUnionJobData.jobLs,
                extraDependentInputLs=subUnionJobData.fileLs,
                inputArgumentOption="--variant")
    
    def concatenateOverlapIntervalsIntoOneVCFSubWorkflow(self, fileBasenamePrefix=None,\
        passingData=None, intervalJobLs=None,
        outputDirJob=None,
        transferOutput=True, job_max_memory=None, walltime=None,
        needBGzipAndTabixJob=True, **keywords):
        """
        Examples:
            realInputVolume = passingData.jobData.file.noOfIndividuals * passingData.jobData.file.noOfLoci
            baseInputVolume = 200*2000000
            #base is 4X coverage in 20Mb region => 120 minutes
            walltime = self.scaleJobWalltimeOrMemoryBasedOnInput(
                realInputVolume=realInputVolume,
                baseInputVolume=baseInputVolume, baseJobPropertyValue=60, \
                minJobPropertyValue=60, maxJobPropertyValue=500).value
            #base is 4X, => 5000M
            job_max_memory = self.scaleJobWalltimeOrMemoryBasedOnInput(
                realInputVolume=realInputVolume,
                baseInputVolume=baseInputVolume, baseJobPropertyValue=2000,
                minJobPropertyValue=2000, maxJobPropertyValue=8000).value
            self.concatenateOverlapIntervalsIntoOneVCFSubWorkflow(
                passingData=passingData, \
                intervalJobLs=[pdata.beagleJob for pdata in mapEachIntervalDataLs],
                outputDirJob=self.beagleReduceDirJob,
                transferOutput=True, job_max_memory=job_max_memory, walltime=walltime,
                **keywords)
        
        #. concatenate overlapping (share some loci) VCFs into one, used in reduceEachVCF
        """
        #ligate vcf job (different segments of a chromosome into one chromosome) for replicate VCFs
        if fileBasenamePrefix is None:
            fileBasenamePrefix = getattr(passingData, 'fileBasenamePrefix', None)
        concatVCFFilename = os.path.join(outputDirJob.folder, '%s.vcf'%(fileBasenamePrefix))
        concatVCFFile = File(concatVCFFilename)
        if needBGzipAndTabixJob:
            transferConcatOutput=False
        else:
            transferConcatOutput = transferOutput
        concatJob = self.addLigateVcfJob(
            executable=self.ligateVcf, \
            ligateVcfExecutableFile=self.ligateVcfExecutableFile, \
            outputFile=concatVCFFile, \
            parentJobLs=[outputDirJob], \
            extraDependentInputLs=None,
            transferOutput=transferConcatOutput, \
            job_max_memory=job_max_memory,
            walltime=walltime/2)
        
        for intervalJob in intervalJobLs:
            #add this output to the union job
            # 2012.6.1 done it through addInputToMergeJob()
            self.addInputToMergeJob(concatJob, inputF=intervalJob.output,
                parentJobLs=[intervalJob],
                extraDependentInputLs=intervalJob.outputLs[1:])
        
        if needBGzipAndTabixJob:
            #bgzip and tabix the trio caller output
            gzipVCFFile = File("%s.gz"%concatVCFFilename)
            bgzip_tabix_job = self.addBGZIP_tabix_Job(
                bgzip_tabix=self.bgzip_tabix_in_reduce,
                parentJob=concatJob,
                inputF=concatJob.output, outputF=gzipVCFFile,
                transferOutput=transferOutput,
                job_max_memory=job_max_memory/4, walltime=walltime/4)
            
            returnData = PassingData(file=bgzip_tabix_job.output,
                vcfFile=bgzip_tabix_job.output, \
                fileLs=[bgzip_tabix_job.output, bgzip_tabix_job.tbi_F],
                job=bgzip_tabix_job, jobLs=[bgzip_tabix_job],
                tbi_F=bgzip_tabix_job.tbi_F)
        else:
            returnData = PassingData(file=concatJob.output,
                vcfFile=concatJob.output,
                fileLs=concatJob.outputLs,
                job=concatJob, jobLs=[concatJob], tbi_F=None)
        return returnData
    
    def concatenateIntervalsIntoOneVCFSubWorkflow(self,
        executable:Transformation=None,
        refFastaFList:List[File]=None, fileBasenamePrefix=None,
        passingData=None, intervalJobLs=None,\
        outputDirJob=None,needBGzipAndTabixJob=True,\
        transferOutput=True, job_max_memory=None, walltime=None,
        **keywords):
        """
        e.g.:
        concatenateVCFJobData = self.concatenateIntervalsIntoOneVCFSubWorkflow(
            refFastaFList=self.newRegisterReferenceData.refFastaFList, \
            fileBasenamePrefix='wholeGenome.%s'%(passingData.fileBasenamePrefix),
            passingData=passingData, \
            intervalJobLs=intervalJobLs,\
            outputDirJob=self.reduceOutputDirJob,
            transferOutput=False, job_max_memory=job_max_memory, walltime=walltime,
            needBGzipAndTabixJob=False)
        #. concatenate non-overlapping VCFs into one, used in reduceEachVCF or reduce()
            assuming all VCFs with same set of samples
        """
        
        if executable is None:
            executable = self.CombineVariantsJavaInReduce
        #ligate vcf job (different segments of a chromosome into one chromosome) for replicate VCFs
        if fileBasenamePrefix is None:
            fileBasenamePrefix = getattr(passingData, 'fileBasenamePrefix', None)
        
        if fileBasenamePrefix is None:
            message = "ERRROR in AbstractVCFWorkflow.concatenateIntervalsIntoOneVCFSubWorkflow(): %s is None."%\
                (fileBasenamePrefix)
            utils.pauseForUserInput(message=message, continueAnswerSet=None, exitType=1)
        
        concatVCFFilename = os.path.join(outputDirJob.folder, '%s.vcf'%(fileBasenamePrefix))
        concatVCFFile = File(concatVCFFilename)
        if needBGzipAndTabixJob:
            transferConcatOutput=False
        else:
            transferConcatOutput = transferOutput
        concatJob = self.addGATKCombineVariantsJob(executable=executable,
            GenomeAnalysisTKJar=None,
            refFastaFList=refFastaFList, inputFileList=None,
            argumentForEachFileInInputFileList="--variant",
            outputFile=concatVCFFile, genotypeMergeOptions='UNSORTED', \
            extraArguments=None, extraArgumentList=['--assumeIdenticalSamples'],
            extraDependentInputLs=None,
            parentJobLs=[outputDirJob], transferOutput=transferConcatOutput,
            job_max_memory=job_max_memory,
            walltime=walltime)
        
        for intervalJob in intervalJobLs:
            self.addInputToMergeJob(concatJob, inputF=intervalJob.output,
                inputArgumentOption="--variant",\
                parentJobLs=[intervalJob],
                extraDependentInputLs=intervalJob.outputLs[1:])
        
        
        if needBGzipAndTabixJob:
            
            #bgzip and tabix the trio caller output
            gzipVCFFile = File("%s.gz"%concatVCFFilename)
            bgzip_tabix_job = self.addBGZIP_tabix_Job(
                bgzip_tabix=self.bgzip_tabix_in_reduce, \
                parentJob=concatJob, \
                inputF=concatJob.output, outputF=gzipVCFFile,
                transferOutput=transferOutput,\
                job_max_memory=job_max_memory/4, walltime=walltime/4)
            
            returnData = self.constructJobDataFromJob(bgzip_tabix_job)
        else:
            returnData = self.constructJobDataFromJob(concatJob)
        return returnData
    
    def reduceManySmallIntervalVCFIntoBigIntervalVCF(self, executable=None,
        registerReferenceData=None, fileBasenamePrefix=None, \
        intervalJobLs=None, outputDirJob=None, bigIntervalSize=10000000,
        transferOutput=False, job_max_memory=None, walltime=None,
        needBGzipAndTabixJob=False,
        **keywords):
        """
        #. this function reduces many small intervals into big intervals
            (capped by bigIntervalSize or chromosome boundary)
        #. each intervalJob should have an intervalData attached to it,
            detailing how big an interval it is covering
        #. assuming all jobs in intervalJobLs are in chromosomal order
        
        intervalJobLs = [jobData.job for jobData in genotypeCallJobData.jobDataLs]
        bigIntervalSize = 40000000	#40Mb
        outputDirJob = self.addMkDirJob(outputDir="%s/bigInterval_%s_VCF"%(downsampleDir, bigIntervalSize))
        bigIntervalGenotypeCallJobData = self.reduceManySmallIntervalVCFIntoBigIntervalVCF(
            executable=self.CombineVariantsJava, \
            registerReferenceData=pdata.registerReferenceData, fileBasenamePrefix="", \
            intervalJobLs=intervalJobLs, outputDirJob=outputDirJob, bigIntervalSize=bigIntervalSize,
            transferOutput=False, job_max_memory=7000, walltime=300, needBGzipAndTabixJob=False)
            
        """
        sys.stderr.write("Reducing %s interval VCFs into VCFs spanning %s bp, current job count=%s ..."%\
            (len(intervalJobLs), bigIntervalSize, self.no_of_jobs))
        returnData = PassingData(jobDataLs=[])
        _intervalJobLs = []
        startIntervalData = None
        #an instance of IntervalData in ParentClass()
        for intervalJob in intervalJobLs:
            intervalData = intervalJob.intervalData
            if startIntervalData is None:
                _intervalJobLs.append(intervalJob)
                startIntervalData = copy.deepcopy(intervalData)
                #not interfering original data
            elif intervalData.chromosome!=startIntervalData.chromosome or \
                (intervalData.span+startIntervalData.span)>bigIntervalSize:
                _filenamePartLs = []
                if fileBasenamePrefix:
                    _filenamePartLs.append(fileBasenamePrefix)
                _filenamePartLs.extend([startIntervalData.chromosome, \
                    startIntervalData.overlapStart, startIntervalData.overlapStop])
                _filenamePartLs = map(str, _filenamePartLs)
                _fileBasenamePrefix = "_".join(_filenamePartLs)
                concateJobData = self.concatenateIntervalsIntoOneVCFSubWorkflow(
                    executable=executable, \
                    refFastaFList=registerReferenceData.refFastaFList, \
                    fileBasenamePrefix=_fileBasenamePrefix, \
                    passingData=None, intervalJobLs=_intervalJobLs,\
                    outputDirJob=outputDirJob,
                    needBGzipAndTabixJob=needBGzipAndTabixJob,\
                    transferOutput=transferOutput, \
                    job_max_memory=job_max_memory, walltime=walltime)
                concateJobData.intervalData = startIntervalData
                returnData.jobDataLs.append(concateJobData)
                
                #reset
                _intervalJobLs = [intervalJob]
                startIntervalData = copy.deepcopy(intervalData)
                #not interfering original data
            else:
                _intervalJobLs.append(intervalJob)
                startIntervalData.unionOneIntervalData(intervalData)
        if _intervalJobLs:	#last remaining intervals jobs
            _fileBasenamePrefix = "%s_%s_%s_%s"%(fileBasenamePrefix, startIntervalData.chromosome, \
                startIntervalData.overlapStart, startIntervalData.overlapStop)
            concateJobData = self.concatenateIntervalsIntoOneVCFSubWorkflow(
                executable=executable, \
                refFastaFList=registerReferenceData.refFastaFList, \
                fileBasenamePrefix=_fileBasenamePrefix, \
                passingData=None, intervalJobLs=_intervalJobLs,\
                outputDirJob=outputDirJob,
                needBGzipAndTabixJob=needBGzipAndTabixJob,\
                transferOutput=transferOutput, job_max_memory=job_max_memory, walltime=walltime)
            concateJobData.intervalData = startIntervalData
            returnData.jobDataLs.append(concateJobData)
        sys.stderr.write(" reduced to %s jobs, current job count=%s.\n"%\
            (len(returnData.jobDataLs), self.no_of_jobs))
        return returnData
    
    def reduceEachChromosome(self, chromosome=None, passingData=None, 
        mapEachVCFDataLs=None, reduceEachVCFDataLs=None,\
        transferOutput=True, \
        **keywords):
        """
        """
        returnData = PassingData(no_of_jobs = 0)
        returnData.jobDataLs = []
        returnData.mapEachVCFDataLs = mapEachVCFDataLs
        returnData.reduceEachVCFDataLs = reduceEachVCFDataLs
        return returnData
        
    def reduceEachVCF(self, chromosome=None, passingData=None,
        mapEachIntervalDataLs=None,\
        transferOutput=True, **keywords):
        """
        """
        return ParentClass.reduceEachInput(self, chromosome=chromosome, \
            passingData=passingData, mapEachIntervalDataLs=mapEachIntervalDataLs,
            transferOutput=transferOutput,\
            **keywords)

    def reduce(self, reduceEachChromosomeDataLs=None, \
        mapEachChromosomeDataLs=None, passingData=None, transferOutput=True, \
        **keywords):
        """
        return each processed-Input job data so that followup workflows could carry out map-reduce
        """
        returnData = PassingData(no_of_jobs = 0)
        returnData.jobDataLs = []
        returnData.mapEachChromosomeDataLs = mapEachChromosomeDataLs
        returnData.reduceEachChromosomeDataLs = reduceEachChromosomeDataLs
        """
        #example to return each processed-Input job data so that followup workflows could carry out map-reduce
        for reduceEachVCFDataLs in passingData.reduceEachVCFDataLsLs:
            if reduceEachVCFDataLs:
                for reduceEachVCFData in reduceEachVCFDataLs:
                    if reduceEachVCFData:
                        returnData.jobDataLs.append(reduceEachVCFData.WHATEVERJobData)
        """
        return returnData
    
    def addSplitVCFSubWorkflow(self, inputVCFData=None, intervalSize=1000000,
        chr2size=None,
        registerReferenceData=None, outputDirJob=None):
        """
        Examples:
            self.addSplitVCFSubWorkflow(inputVCFData=None, intervalSize=1000000, outputDirJob=None)
        2013.09.19
            split each VCF into non-overlapping sub-VCFs
        """
        if chr2size is None:
            chr2size = self.chr2size
        if registerReferenceData is None:
            registerReferenceData = self.registerReferenceData
        sys.stderr.write("Splitting %s VCFs into sub-VCFs with intervalSize=%s, jobCount=%s ..."%
                        (len(inputVCFData.jobDataLs), intervalSize, self.no_of_jobs))
        returnData = PassingData(jobDataLs=[])
        for jobData in inputVCFData.jobDataLs:
            filenameParseData = Genome.parseChrStartStopFromFilename(
                filename=jobData.file.name, chr2size=chr2size)
            noOfIntervals = max(1, utils.getNoOfUnitsNeededToCoverN(
                N=filenameParseData.stop-filenameParseData.start+1, \
                s=intervalSize, o=0)-1)
            if noOfIntervals==1:	#no splitting
                returnData.jobDataLs.append(jobData)
            else:
                for i in range(noOfIntervals):
                    _start = filenameParseData.start + i*intervalSize
                    if i==noOfIntervals-1:
                        _stop = filenameParseData.stop
                    else:
                        _stop = min(_start + intervalSize-1, filenameParseData.stop)
                    fileBasenamePrefix = "%s"%(utils.getFileBasenamePrefixFromPath(jobData.file.name))
                    outputF = File(os.path.join(outputDirJob.output, "%s_%s_%s_%s.vcf"%(
                        fileBasenamePrefix, filenameParseData.chromosome, _start, _stop)))
                    selectVCFJob = self.addSelectVariantsJob(
                        SelectVariantsJava=self.SelectVariantsJava, \
                        inputF=jobData.file, outputF=outputF, \
                        interval="%s:%s-%s"%(filenameParseData.chromosome, _start, _stop),\
                        refFastaFList=registerReferenceData.refFastaFList, \
                        extraArguments=None, extraArgumentList=None,
                        parentJobLs=[outputDirJob] + jobData.jobLs,
                        extraDependentInputLs=jobData.fileLs[1:], transferOutput=False, \
                        job_max_memory=2000, walltime=None)
                    returnData.jobDataLs.append(self.constructJobDataFromJob(selectVCFJob))
        sys.stderr.write("jobCount=%s.\n"%(self.no_of_jobs))
        return returnData
    
    def addAllJobs(self, inputVCFData=None, chr2IntervalDataLs=None, \
        GenomeAnalysisTKJar=None, samtools=None, \
        CreateSequenceDictionaryJava=None, CreateSequenceDictionaryJar=None, \
        BuildBamIndexFilesJava=None, BuildBamIndexJar=None,\
        mv=None, \
        registerReferenceData=None, \
        needFastaIndexJob=False, needFastaDictJob=False, \
        data_dir=None, no_of_gatk_threads = 1, \
        intervalSize=3000, intervalOverlapSize=0, \
        outputDirPrefix="", passingData=None, \
        transferOutput=True, job_max_memory=2000, **keywords):
        """
        2013.06.14 bugfix regarding noOfUnits, which was all inferred from one file
        2012.7.26
            architect of the whole map-reduce framework
        """
        chr2jobDataLs = {}
        for jobData in inputVCFData.jobDataLs:
            chromosome = self.getChrFromFname(os.path.basename(jobData.file.name))
            if chromosome not in chr2jobDataLs:
                chr2jobDataLs[chromosome] = []
            chr2jobDataLs[chromosome].append(jobData)
        
        sys.stderr.write("Adding jobs for %s chromosomes/contigs of %s VCF files... \n"%(
            len(chr2jobDataLs), len(inputVCFData.jobDataLs)))
        if getattr(registerReferenceData, 'refFastaFList', None):
            refFastaFList = registerReferenceData.refFastaFList
        else:
            refFastaFList = None
        if refFastaFList:
            refFastaF = refFastaFList[0]
        else:
            refFastaF = None
        
        if needFastaDictJob or getattr(registerReferenceData, 'needPicardFastaDictJob', None):
            fastaDictJob = self.addRefFastaDictJob(
                CreateSequenceDictionaryJava=CreateSequenceDictionaryJava, \
                CreateSequenceDictionaryJar=self.CreateSequenceDictionaryJar,\
                refFastaF=refFastaF)
            refFastaDictF = fastaDictJob.refFastaDictF
        else:
            fastaDictJob = None
            refFastaDictF = getattr(registerReferenceData, 'refPicardFastaDictF', None)
        
        if needFastaIndexJob or getattr(registerReferenceData, 'needSAMtoolsFastaIndexJob', None):
            fastaIndexJob = self.addRefFastaFaiIndexJob(samtools=samtools, refFastaF=refFastaF)
            refFastaIndexF = fastaIndexJob.refFastaIndexF
        else:
            fastaIndexJob = None
            refFastaIndexF = getattr(registerReferenceData, 'refSAMtoolsFastaIndexF', None)
        
        returnData = PassingData()
        returnData.jobDataLs = []
        
        #2012.9.22 
        # 	mapEachAlignmentDataLs is never reset.
        #	mapEachChromosomeDataLs is reset upon new alignment
        #	mapEachIntervalDataLs is reset upon each new chromosome
        #	all reduce lists never get reset.
        #	fileBasenamePrefix is the prefix of input file's basename,
        #       to be used for temporary output files in reduceEachVCF()
        #       but not for output files in mapEachInterval()
        passingData = PassingData(\
                    fileBasenamePrefix=None, \
                    chromosome=None, \
                    
                    outputDirPrefix=outputDirPrefix, \
                    intervalFileBasenamePrefix=None,\
                    
                    refFastaFList=refFastaFList, \
                    registerReferenceData=registerReferenceData, \
                    refFastaF=refFastaF,\
                    
                    fastaDictJob = fastaDictJob,\
                    refFastaDictF = refFastaDictF,\
                    fastaIndexJob = fastaIndexJob,\
                    refFastaIndexF = refFastaIndexF,\
                    
                    intervalOverlapSize =intervalOverlapSize, intervalSize=intervalSize,\
                    jobData=None,\
                    VCFJobData=None,\
                    splitVCFFile=None,\
                    intervalDataLs=None,\
                    preReduceReturnData=None,\
                    
                    mapEachIntervalData=None,\
                    mapEachIntervalDataLs=None,\
                    mapEachIntervalDataLsLs=[],\
                    mapEachVCFData=None,\
                    mapEachVCFDataLs=None,\
                    mapEachVCFDataLsLs=[],\
                    mapEachChromosomeData=None, \
                    mapEachChromosomeDataLs=[], \
                    
                    reduceEachVCFData=None,\
                    reduceEachVCFDataLs=None,\
                    reduceEachVCFDataLsLs=[],\
                    
                    reduceEachChromosomeData=None,\
                    reduceEachChromosomeDataLs=[],\
                    
                    chr2jobDataLs = chr2jobDataLs,\
                    )
        # mapEachIntervalDataLsLs is list of mapEachIntervalDataLs by each VCF file.
        # mapEachVCFDataLsLs is list of mapEachVCFDataLs by each chromosome
        # reduceEachVCFDataLsLs is list of reduceEachVCFDataLs by each chromosome
        
        preReduceReturnData = self.preReduce(outputDirPrefix=outputDirPrefix, \
                                    passingData=passingData, transferOutput=True,\
                                    **keywords)
        passingData.preReduceReturnData = preReduceReturnData
        
        #gzip folder jobs (to avoid repeatedly creating the same folder
        gzipReduceEachVCFFolderJob = None
        gzipReduceEachChromosomeFolderJob = None
        gzipReduceFolderJob = None
        gzipPreReduceFolderJob = None
        no_of_vcf_files = 0
        for chromosome, jobDataLs in chr2jobDataLs.items():
            passingData.chromosome = chromosome
            mapEachChromosomeData = self.mapEachChromosome(chromosome=chromosome, \
                                        passingData=passingData, \
                                        transferOutput=False, **keywords)
            passingData.mapEachChromosomeData = mapEachChromosomeData
            passingData.mapEachChromosomeDataLs.append(mapEachChromosomeData)
            
            passingData.mapEachVCFDataLsLs.append([])
            #the last one from the double list is the current one
            passingData.mapEachVCFDataLs = passingData.mapEachVCFDataLsLs[-1]
            
            passingData.reduceEachVCFDataLsLs.append([])
            passingData.reduceEachVCFDataLs = passingData.reduceEachVCFDataLsLs[-1]
            
            #2013.06.20
            if chr2IntervalDataLs:
                passingData.intervalDataLs = chr2IntervalDataLs.get(chromosome)
            else:
                pass
            for i in range(len(jobDataLs)):	#each input job is one VCF 
                jobData = jobDataLs[i]
                passingData.jobData = jobData
                passingData.VCFJobData = jobData
                
                VCFFile = jobData.vcfFile
                inputFBaseName = os.path.basename(VCFFile.name)
                commonPrefix = inputFBaseName.split('.')[0]
                
                passingData.fileBasenamePrefix = commonPrefix
                
                no_of_vcf_files += 1
                if no_of_vcf_files%10==0:
                    sys.stderr.write("%s\t%s VCFs."%('\x08'*40, no_of_vcf_files))
                if self.mapReduceType==1:
                    # type 1: split VCF with fixed number of sites
                    mapEachVCFData = self.mapEachVCF(VCFJobData=jobData, passingData=passingData, \
                        transferOutput=False, **keywords)
                    passingData.mapEachVCFData = mapEachVCFData
                    passingData.mapEachVCFDataLs.append(mapEachVCFData)
                    
                    passingData.mapEachIntervalDataLsLs.append([])
                    passingData.mapEachIntervalDataLs = passingData.mapEachIntervalDataLsLs[-1]
                    
                    splitVCFJob = mapEachVCFData.splitVCFJob
                    #noOfUnits = max(1, utils.getNoOfUnitsNeededToCoverN(N=jobData.file.noOfLoci,
                    #  s=intervalSize, o=intervalOverlapSize)-1)
                    noOfUnits = splitVCFJob.noOfUnits
                    jobData.file.noOfUnits = noOfUnits
                    for unitNumber in range(1, noOfUnits+1):
                        splitVCFFile = getattr(splitVCFJob, 'unit%sFile'%(unitNumber), None)
                        if splitVCFFile is not None:
                            passingData.splitVCFFile = splitVCFFile
                            passingData.unitNumber = unitNumber
                            passingData.intervalFileBasenamePrefix = '%s_%s_splitVCF_u%s'%(
                                chromosome, commonPrefix, unitNumber)
                            passingData.noOfIndividuals = getattr(jobData.file, 'noOfIndividuals', None)
                            passingData.span = self.intervalSize + self.intervalOverlapSize*2 
                            #2013.06.19 for memory/walltime gauging
                            mapEachIntervalData = self.mapEachInterval(\
                                VCFJobData=PassingData(file=splitVCFFile,
                                    vcfFile=splitVCFFile, fileLs=[splitVCFFile], \
                                    job=splitVCFJob, jobLs=[splitVCFJob], tbi_F=None), \
                                chromosome=chromosome,intervalData=None,\
                                mapEachChromosomeData=mapEachChromosomeData, \
                                passingData=passingData, transferOutput=False, \
                                **keywords)
                            passingData.mapEachIntervalData = mapEachIntervalData
                            passingData.mapEachIntervalDataLs.append(mapEachIntervalData)
                            
                            linkMapToReduceData = self.linkMapToReduce(
                                mapEachIntervalData=mapEachIntervalData, \
                                preReduceReturnData=preReduceReturnData, \
                                passingData=passingData, \
                                **keywords)
                else:# type 2: SelectVariants from VCF with fixed-size windows
                    if not chr2IntervalDataLs:
                        logging.error(f"mapReduceType={self.mapReduceType}, "
                            f"but chr2IntervalDataLs ({repr(chr2IntervalDataLs)}) is nothing.")
                        sys.exit(2)
                    if not chr2IntervalDataLs.get(chromosome):
                        logging.error(f"chromosome {chromosome} is not in chr2IntervalDataLs "
                            f"(mapReduceType={self.mapReduceType}) while "
                            f"VCFFile {jobData.file.name} is from this chromosome.")
                        sys.exit(3)
                    for intervalData in chr2IntervalDataLs.get(chromosome):
                        mapEachIntervalData = self.mapEachInterval(
                            VCFJobData=jobData,
                            chromosome=chromosome, intervalData=intervalData,
                            mapEachChromosomeData=mapEachChromosomeData,
                            passingData=passingData, transferOutput=False,
                            **keywords)
                        passingData.mapEachIntervalData = mapEachIntervalData
                        passingData.mapEachIntervalDataLs.append(mapEachIntervalData)
                        
                        linkMapToReduceData = self.linkMapToReduce(
                            mapEachIntervalData=mapEachIntervalData,
                            preReduceReturnData=preReduceReturnData,
                            passingData=passingData,
                            **keywords)
                reduceEachVCFData = self.reduceEachVCF(chromosome=chromosome,
                    passingData=passingData,
                    mapEachIntervalDataLs=passingData.mapEachIntervalDataLs,
                    transferOutput=False, data_dir=data_dir,
                    **keywords)
                passingData.reduceEachVCFData = reduceEachVCFData
                passingData.reduceEachVCFDataLs.append(reduceEachVCFData)
                
                gzipReduceEachVCFData = self.addGzipSubWorkflow(
                    inputData=reduceEachVCFData,
                    outputDirPrefix="%sReduceEachVCF"%(outputDirPrefix),
                    topOutputDirJob=gzipReduceEachVCFFolderJob,
                    transferOutput=transferOutput,
                    report=False)
                gzipReduceEachVCFFolderJob = gzipReduceEachVCFData.topOutputDirJob
            reduceEachChromosomeData = self.reduceEachChromosome(
                chromosome=chromosome, passingData=passingData,
                mapEachVCFDataLs=passingData.mapEachVCFDataLs,
                reduceEachVCFDataLs=passingData.reduceEachVCFDataLs,
                transferOutput=False, data_dir=data_dir,
                **keywords)
            passingData.reduceEachChromosomeData = reduceEachChromosomeData
            passingData.reduceEachChromosomeDataLs.append(reduceEachChromosomeData)
            
            gzipReduceEachChromosomeData = self.addGzipSubWorkflow(\
                    inputData=reduceEachChromosomeData, transferOutput=transferOutput,\
                    outputDirPrefix="%sReduceEachChromosome"%(outputDirPrefix), \
                    topOutputDirJob=gzipReduceEachChromosomeFolderJob, report=False)
            gzipReduceEachChromosomeFolderJob = gzipReduceEachChromosomeData.topOutputDirJob
            
        reduceReturnData = self.reduce(passingData=passingData, transferOutput=False,
            mapEachChromosomeDataLs=passingData.mapEachVCFDataLs,\
            reduceEachChromosomeDataLs=passingData.reduceEachChromosomeDataLs,\
            **keywords)
        passingData.reduceReturnData = reduceReturnData
        
        if self.needGzipPreReduceReturnData:
            gzipPreReduceReturnData = self.addGzipSubWorkflow(
                inputData=preReduceReturnData, transferOutput=transferOutput,
                outputDirPrefix="%sPreReduce"%(outputDirPrefix),
                topOutputDirJob= gzipPreReduceFolderJob, report=False)
            gzipPreReduceFolderJob = gzipPreReduceReturnData.topOutputDirJob
        
        if self.needGzipReduceReturnData:
            gzipReduceReturnData = self.addGzipSubWorkflow(
                inputData=reduceReturnData, transferOutput=transferOutput,
                outputDirPrefix="%sReduce"%(outputDirPrefix), \
                topOutputDirJob=gzipReduceFolderJob, report=False)
            gzipReduceFolderJob = gzipReduceReturnData.topOutputDirJob
        
        sys.stderr.write("\n %s%s VCF files.\n"%('\x08'*40, no_of_vcf_files))
        sys.stderr.write("%s jobs.\n"%(self.no_of_jobs))
        return reduceReturnData
    
    def setup_run(self):
        """
        2013.06.11 added firstVCFJobData in return
            assign all returned data to self, rather than pdata (pdata has become self)
        2013.04.07 wrap all standard pre-run() related functions into this function.
            setting up for run(), called by run()
        """
        pdata = ParentClass.setup_run(self)
        
        #self.chr2size = {}
        #self.chr2size = set(['Contig149'])	#temporary when testing Contig149
        #self.chr2size = set(['1MbBAC'])
        # #temporary when testing the 1Mb-BAC (formerly vervet_path2)
        if self.needSplitChrIntervalData:
            #2013.06.21 defined in ParentClass.__init__()
            chr2IntervalDataLs = self.getChr2IntervalDataLsBySplitChrSize(
                chr2size=self.chr2size,
                intervalSize=self.intervalSize, \
                intervalOverlapSize=self.intervalOverlapSize)
        else:
            chr2IntervalDataLs = None
        inputData = None
        firstVCFJobData = None
        if getattr(self, 'input_path', None):
            inputData = self.registerFilesOfInputDir(
                inputDir=self.input_path,
                input_site_handler=self.input_site_handler,
                checkEmptyVCFByReading=self.checkEmptyVCFByReading,
                pegasusFolderName=self.pegasusFolderName,\
                maxContigID=self.maxContigID, \
                minContigID=self.minContigID,\
                db_main=getattr(self, 'db_main', None), \
                needToKnowNoOfLoci=getattr(self, 'needToKnowNoOfLoci', True),
                minNoOfLociInVCF=getattr(self, 'minNoOfLociInVCF', 10))
            if inputData and inputData.jobDataLs:
                firstVCFJobData = inputData.jobDataLs[0]
                firstVCFFile = firstVCFJobData.file
                print(f"\t VCF file {firstVCFFile} is chosen as an example VCF "
                    "for any job that needs a random VCF file.", flush=True)
        
        
        self.inputData = inputData
        self.chr2IntervalDataLs = chr2IntervalDataLs
        self.firstVCFJobData = firstVCFJobData
        #self.firstVCFFile = firstVCFFile
        return self
    
    def run(self):
        """
        """
        pdata = self.setup_run()
        inputData=pdata.inputData
        if len(inputData.jobDataLs)<=0:
            logging.error(f"No VCF files in this folder, {self.input_path}.")
            sys.exit(0)
                
        self.addAllJobs(inputVCFData=inputData, inputData=inputData,
                chr2IntervalDataLs=self.chr2IntervalDataLs,
                samtools=self.samtools,
                GenomeAnalysisTKJar=self.GenomeAnalysisTKJar,
                CreateSequenceDictionaryJava=self.CreateSequenceDictionaryJava,
                CreateSequenceDictionaryJar=self.CreateSequenceDictionaryJar,
                BuildBamIndexFilesJava=self.BuildBamIndexFilesJava,
                BuildBamIndexJar=self.BuildBamIndexJar,
                mv=self.mv,
                registerReferenceData=pdata.registerReferenceData,
                needFastaIndexJob=getattr(self, 'needFastaIndexJob',False),
                needFastaDictJob=getattr(self, 'needFastaDictJob', False),
                data_dir=self.data_dir, no_of_gatk_threads = 1,\
                intervalSize=self.intervalSize,
                intervalOverlapSize=self.intervalOverlapSize,
                outputDirPrefix=self.pegasusFolderName, transferOutput=True)
        
        self.end_run()



#!/usr/bin/env python3
"""
Description:
    20120509 Parent of other VCF workflows.

Examples:
    #2012.5.11 convert alignment read group (sample id) into UCLAID
    %s -I FilterVCF_VRC_SK_Nevis_FilteredSeq_top1000Contigs.2012.5.6_trioCaller.2012.5.8T21.42/trioCaller_vcftoolsFilter/ 
        -o dags/SampleIDInUCLAID_FilterVCF_VRC_SK_Nevis_FilteredSeq_top1000Contigs.2012.5.6_trioCaller.2012.5.8.xml 
        -u yh -y4 -l hcondor -j hcondor  -z localhost
        -e /u/home/eeskin/polyacti/
        -t /u/home/eeskin/polyacti/NetworkData/vervet/db/
        -D /u/home/eeskin/polyacti/NetworkData/vervet/db/ 
    
    # 2012.5.10 subset + convert-2-plink.
    # run on hoffman2 condor, minMAC=1 (-n 1), minMAF=0.1 (-f 0.1), maxSNPMissingRate=0 (-L 0)
    #    (turn on checkEmptyVCFByReading, --checkEmptyVCFByReading)
    %s -I FilterVCF_VRC_SK_Nevis_FilteredSeq_top1000Contigs.2012.5.6_trioCaller.2012.5.8T21.42/trioCaller_vcftoolsFilter/
        -o dags/SubsetTo36RNASamplesAndPlink_FilterVCF_VRC_SK_Nevis_FilteredSeq_top1000Contigs.2012.5.6_trioCaller.2012.5.8.xml
        -i ~/script/vervet/data/RNADevelopment_eQTL/36monkeys.phenotypes.txt
        -w ~/script/vervet/data/RNADevelopment_eQTL/36monkeys.inAlignmentReadGroup.tsv
        -n1 -f 0.1 -L 0 -y3 --checkEmptyVCFByReading
        -l hcondor -j hcondor  -u yh -z localhost --needSSHDBTunnel
        -e /u/home/eeskin/polyacti/
        -D /u/home/eeskin/polyacti/NetworkData/vervet/db/
        -t /u/home/eeskin/polyacti/NetworkData/vervet/db/
    
    # 2012.7.16 convert a folder of VCF files into plink, need the db tunnel (--needSSHDBTunnel) for output pedigree in tfam
    # "-V 90 -x 100" are used to restrict contig IDs between 90 and 100.
    %s -I FilterVCF_VRC_SK_Nevis_FilteredSeq_top1000Contigs.MAC10.MAF.05_trioCaller/trioCaller_vcftoolsFilter/ 
        -o dags/ToPlinkFilterVCF_VRC_SK_Nevis_FilteredSeq_top1000Contigs.MAC10.MAF.05.xml
        -y 2 --checkEmptyVCFByReading
        -l condorpool -j condorpool
        -u yh -z uclaOffice  -C 4 --needSSHDBTunnel
        #-V 90 -x 100 
    
    # 2012.7.25 calculate haplotype distance & majority call support stats
    %s -I AlignmentToTrioCall_VRC_FilteredSeq.2012.7.21T0248_VCFWithReplicates/
        -o dags/GetReplicateHaplotypeStat_TrioCall_VRC_FilteredSeq.2012.7.21T0248_VCFWithReplicats.xml
        -y 5 --checkEmptyVCFByReading -l condorpool -j condorpool -u yh -z uclaOffice  -C 1 -a 524
    
    # 2012.8.20 convert method 16 to yu format (-y 6 works for generic VCF, -y 7 adds sample ID conversion first)
    %s -I ~/NetworkData/vervet/db/genotype_file/method_16/
        -o dags/VCF2YuFormat/VCF2YuFormat_Method16.xml
        -y 7 --checkEmptyVCFByReading  -l hcondor -j hcondor  -u yh -z localhost --needSSHDBTunnel -C 2
        -D /u/home/eeskin/polyacti/NetworkData/vervet/db/  -t /u/home/eeskin/polyacti/NetworkData/vervet/db/

    # 2012.8.30 combine multiple VCF into one
    # -s .. is optional. if given, the combined VCF will be added into db.
    %s -I ~/NetworkData/vervet/db/genotype_file/method_10/ -o dags/GenericVCFWorkflow/MultiVCF2OneFile_Method10.xml
        -y 9  -l hcondor -j hcondor -u yh -z localhost --needSSHDBTunnel -C 1
        -s 16HCSAMtoolsMinDP1_2FoldDepth_minMAC8_maxSNPMissing0
        -D /u/home/eeskin/polyacti/NetworkData/vervet/db/ -t /u/home/eeskin/polyacti/NetworkData/vervet/db/
    
"""
import sys, os
__doc__ = __doc__%(sys.argv[0], sys.argv[0], sys.argv[0], sys.argv[0], \
    sys.argv[0], sys.argv[0])

import csv, copy
from palos import PassingData, \
    figureOutDelimiter, getColName2IndexFromHeader, utils
from palos.ngs.io.VCFFile import VCFFile
from pegaflow.api import File
import pegaflow
#from palos.pegasus.AbstractVCFWorkflow import AbstractVCFWorkflow
from palos.db import SunsetDB
from . AbstractNGSWorkflow import AbstractNGSWorkflow

ParentClass = AbstractNGSWorkflow
class GenericVCFWorkflow(ParentClass):
    __doc__ = __doc__
    option_default_dict = copy.deepcopy(ParentClass.option_default_dict)
    option_default_dict.update({
        ('individualUCLAIDFname', 0, ): [None, 'i', 1, 
            'a file containing individual ucla_id in each row. one column with header UCLAID. ', ],\
        ('vcfSampleIDFname', 0, ): [None, 'w', 1, 
            'a file containing the sample ID (a composite ID including ucla_id) each row. '
            'any samples not in this file will be removed in subset VCF run_type (1, 3)'
            'You can also use individualUCLAIDFname to specify the sample IDs (UCLAID). '
            'Their composite IDs will be inferred from individualUCLAIDFname + first VCF file header.', ],
        ('vcf2Dir', 0, ): ['', '', 1, 
            'the 2nd input folder that contains vcf or vcf.gz files.', ],
        ('vcfSubsetPath', 1, ): ["bin/vcftools/vcf-subset", '', 1, 
            'path to the vcf-subset program'],
        ('run_type', 1, int): [1, 'y', 1, 'which run_type to run. \
            1: subset VCF based on input file containing sample IDs;\
            2: convert to plink format; \
            3: subset + convert-2-plink. MAC & MAF & maxSNPMissingRate applied in the convert-to-plink step.\
            4: ConvertAlignmentReadGroup2UCLAIDInVCF jobs.\
            5: addMergeVCFReplicateHaplotypesJobs to get haplotype distance & majority call support stats.\
            6: VCF2YuFormatJobs, \
            7: ConvertAlignmentReadGroup2UCLAIDInVCF + VCF2YuFormatJobs, \
            8: ConvertAlignmentReadGroup2UCLAIDInVCF + convert to plink format, \
            9: combine all single-chromosome VCF into one. \
            ?: Combine VCF files from two input folder, chr by chr. (not done yet. check CheckTwoVCFOverlapPipeline.py for howto)', ],\
        ("minMAC", 0, int): [None, 'n', 1, 'minimum MinorAlleleCount (by chromosome)'],\
        ("minMAF", 0, float): [None, 'f', 1, 
            'minimum MinorAlleleFrequency (by chromosome)'],\
        ("maxSNPMissingRate", 0, float): [1.0, 'L', 1, 
            'maximum SNP missing rate in one vcf (denominator is #chromosomes)'],\
        ('genotypeMethodShortName', 0, ):[None, 's', 1, 
            'column short_name of GenotypeMethod table,'
            'will be created if not present in db.'
            'for run_type 9, if given the file would be added into db.'],\
        })

    def __init__(self,
        vcfSubsetPath="bin/vcftools/vcf-subset",
        plinkPath="bin/plink",
        **keywords):
        """
        """
        self.vcfSubsetPath = vcfSubsetPath
        self.plinkPath = plinkPath
        # Set before ParentClass.__init__()
        self.pathToInsertHomePathList.extend(['vcfSubsetPath', 'plinkPath'])
        ParentClass.__init__(self, **keywords)
    
    def addPlinkJob(self, executable=None, inputFileList=None,
        parentPlinkJob=None,\
        tpedFile=None, tfamFile=None,\
        pedFile=None, famFile=None, mapFile=None,
        bedFile=None, bimFile=None,
        inputFnamePrefix=None, inputOption='--file', \
        outputFnamePrefix=None, outputOption='--out',\
        makeBED=False, calculateMendelError=False, checkSex=False,
        LDPruneWindowSize=100, LDPruneWindowShiftSize=5,
        LDPruneByPairwiseR2=False, LDPruneMinR2=0.1,\
        LDPruneByRegression=False, LDPruneMinVarianceInflationFactor=2,
        estimatePairwiseGenomeWideIBD=False,
        estimatePairwiseGenomeWideIBDFreqFile=None,
        extractSNPFile=None, recodeOutput=False,
        recodeTransposeOutput=False,
        estimateAlleFrequency=False,
        mergeListFile=None,\
        parentJobLs=None, extraDependentInputLs=None,
        transferOutput=False,
        extraArguments=None, extraArgumentList=None,
        extraOutputLs =None,
        job_max_memory=2000, **keywords):
        """
        i.e.

        bedFnamePrefix = os.path.join(topOutputDir, '%s_bed'%(commonPrefix))
        convertSingleTPED2BEDJob = self.addPlinkJob(executable=self.plink, \
            inputFileList=[],
            tpedFile=modifyTPEDJob.output, tfamFile=tfamJob.tfamFile,\
            outputFnamePrefix=bedFnamePrefix, outputOption='--out',\
            makeBED=True, \
            extraDependentInputLs=None, transferOutput=transferOutput, \
            extraArguments=None, job_max_memory=2000,\
            parentJobLs = convertSingleTPED2BEDParentJobLs)


        convertMergedTPED2BEDJob = self.addPlinkJob(executable=self.plink,
                inputFileList=[tpedFileMergeJob.output, tfamJob.tfamFile], \
                inputFnamePrefix=mergedPlinkFnamePrefix, inputOption='--tfile', \
            outputFnamePrefix=mergedPlinkBEDFnamePrefix, outputOption='--out',\
            makeBED=True, \
            extraDependentInputLs=None, transferOutput=transferOutput, \
            extraArguments=None, job_max_memory=2000, parentJobLs=[mergedOutputDirJob, 
                tpedFileMergeJob, tfamJob])

        mendelFnamePrefix = os.path.join(setupData.mapDirJob.output, '%s'%(commonPrefix))
        if inputJob.output.name[-4:]=='tped':
            inputJob.tfamFile = tfamJob.tfamFile
        plinkMendelJob = self.addPlinkJob(executable=self.plink, \
                parentPlinkJob=inputJob,\
                outputFnamePrefix=mendelFnamePrefix, outputOption='--out',\
                calculateMendelError=True, \
                extraDependentInputLs=None,
                transferOutput=transferOneContigPlinkOutput, \
                extraArguments=None, job_max_memory=2000,\
                parentJobLs =[setupData.mapDirJob, tfamJob]+ jobData.jobLs)

        for plink mendel, LD-prune and other jobs, add extraArguments="--allow-no-sex"
            to include individuals without sex

        2013.07.25 added parentPlinkJob (returned from this function),
            and parse input from that job
        2013.07.24 added argument recodeTransposeOutput (--recode --transpose)
        2012.8.28
            add argument
                estimateAlleFrequency, estimate frequency of input file.
                     "--nonfounders" could be added as well.
                estimatePairwiseGenomeWideIBDFreqFile, is the file from which 
                    IBD check could draw frequency (rather than estimate from founders)

        2012.8.9
            inputFileList is a list of pegasus Files (.ped, .fam, or .tped,
                 .tfam, etc.) or could be supplied individually.

            inputOption could be, "--file" for .ped .map ; "--tfile" for .tped,
                 .tfam; or '--bfile' for .bed, .fam, .bim

            if extractSNPFile or mergeListFile is given, either recodeOutput
                 or makeBED have to be on. otherwise, no output.
            http://pngu.mgh.harvard.edu/~purcell/plink/index.shtml
        """
        if extraDependentInputLs is None:
            extraDependentInputLs = []
        if inputFileList:
            extraDependentInputLs.extend(inputFileList)

        if extraArgumentList is None:
            extraArgumentList = []
        if extraOutputLs is None:
            extraOutputLs = []
        key2ObjectForJob = {}

        if parentPlinkJob:
            if bedFile is None:
                bedFile = getattr(parentPlinkJob, 'bedFile', None)
            if famFile is None:
                famFile = getattr(parentPlinkJob, 'famFile', None)
            if bimFile is None:
                bimFile = getattr(parentPlinkJob, 'bimFile', None)
            if tpedFile is None:
                tpedFile = getattr(parentPlinkJob, 'tpedFile', None)
            if tfamFile is None:
                tfamFile = getattr(parentPlinkJob, 'tfamFile', None)
            if mapFile is None:
                mapFile = getattr(parentPlinkJob, 'mapFile', None)
            if pedFile is None:
                pedFile = getattr(parentPlinkJob, 'pedFile', None)
            if famFile is None:
                famFile = getattr(parentPlinkJob, 'famFile', None)

        if inputOption and inputFnamePrefix:
            extraArgumentList.extend([inputOption, inputFnamePrefix])
        if tpedFile:
            extraDependentInputLs.append(tpedFile)
            extraArgumentList.extend(["--tped", tpedFile])
        if tfamFile:
            extraDependentInputLs.append(tfamFile)
            extraArgumentList.extend(["--tfam", tfamFile])
        if pedFile:
            extraDependentInputLs.append(pedFile)
            extraArgumentList.extend(["--ped", pedFile])
        if famFile:
            extraDependentInputLs.append(famFile)
            extraArgumentList.extend(["--fam", famFile])
        if mapFile:
            extraDependentInputLs.append(mapFile)
            extraArgumentList.extend(["--map", mapFile])
        if bedFile:
            extraDependentInputLs.append(bedFile)
            extraArgumentList.extend(["--bed", bedFile])
        if bimFile:
            extraDependentInputLs.append(bimFile)
            extraArgumentList.extend(["--bim", bimFile])

        if outputFnamePrefix and outputOption:
            extraArgumentList.extend([outputOption, outputFnamePrefix])
        else:
            outputFnamePrefix = 'plink'

        suffixAndNameTupleList = []
            # a list of tuples , in each tuple, 1st element is the suffix.
            #  2nd element is the proper name of the suffix.
            #job.$nameFile will be the way to access the file.
            #if 2nd element (name) is missing, suffix[1:].replace('.', '_')
            #  is the name (dot replaced by _)
        if makeBED:
            extraArgumentList.append('--make-bed')
            suffixAndNameTupleList.extend([['.bed',], ('.fam',), ['.bim',]])
            #, binary map file, is excluded for now
        if calculateMendelError:
            extraArgumentList.append('--mendel')
            suffixAndNameTupleList.extend([('.mendel',), ('.imendel',), 
                ('.fmendel',), ('.lmendel',)])
            #its output is not tab-delimited. rather it's space (multi) delimited.
        if checkSex:
            extraArgumentList.append('--check-sex')
            suffixAndNameTupleList.extend([('.sexcheck',), ('.hh', )])
            #.sexcheck file is accessible as job.sexcheckFile.
            #.hh is heterozygous haplotype genotypes
        if LDPruneByPairwiseR2:
            extraArgumentList.append('--indep-pairwise %s %s %s'%(\
                LDPruneWindowSize, LDPruneWindowShiftSize, LDPruneMinR2))
            suffixAndNameTupleList.extend([('.prune.in',), ('.prune.out',)])
            #".prune.in" is accessible as job.prune_inFile
        if LDPruneByRegression:
            extraArgumentList.append('--indep %s %s %s'%(LDPruneWindowSize,\
                LDPruneWindowShiftSize, LDPruneMinVarianceInflationFactor))
            suffixAndNameTupleList.extend([('.prune.in',), ('.prune.out',)])
            #".prune.in" is accessible as job.prune_inFile
        if estimatePairwiseGenomeWideIBD:
            extraArgumentList.append('--genome')
            suffixAndNameTupleList.extend([('.genome',)])
            #.genome is accessible as job.genomeFile
            if estimatePairwiseGenomeWideIBDFreqFile:
                extraArgumentList.extend(['--read-freq', \
                    estimatePairwiseGenomeWideIBDFreqFile])
                extraDependentInputLs.append(estimatePairwiseGenomeWideIBDFreqFile)
        if extractSNPFile:
            extraArgumentList.extend(['--extract', extractSNPFile])
            extraDependentInputLs.append(extractSNPFile)
        if recodeOutput:
            extraArgumentList.extend(['--recode',])
            suffixAndNameTupleList.extend([('.ped',), ('.map',)])
        if recodeTransposeOutput:
            extraArgumentList.extend(['--recode', "--transpose"])
            suffixAndNameTupleList.extend([('.tped',), ('.tfam',)])
        if estimateAlleFrequency:
            extraArgumentList.append('--freq')
            suffixAndNameTupleList.extend([('.frq',)])

        if mergeListFile:
            extraArgumentList.extend(['--merge-list', mergeListFile])
            extraDependentInputLs.append(mergeListFile)
        if extraArguments:
            extraArgumentList.append(extraArguments)


        self.setupMoreOutputAccordingToSuffixAndNameTupleList(
            outputFnamePrefix=outputFnamePrefix, \
            suffixAndNameTupleList=suffixAndNameTupleList, \
            extraOutputLs=extraOutputLs, key2ObjectForJob=key2ObjectForJob)
        #2013.07.24 add it in the end
        #2012.8.10 left in the folder dying
        logFile = File('%s.log'%(outputFnamePrefix))
        extraOutputLs.append(logFile)

        job= self.addGenericJob(executable=executable, inputFile=None,
            outputFile=None, \
            parentJobLs=parentJobLs,
            extraDependentInputLs=extraDependentInputLs,
            extraOutputLs=extraOutputLs,
            transferOutput=transferOutput,
            extraArgumentList=extraArgumentList,
            key2ObjectForJob=key2ObjectForJob, job_max_memory=job_max_memory,
            **keywords)
        return job
    
    def addVCF2PlinkJobs(self, inputData=None, db_main=None,
        minMAC=None, minMAF=None,
        maxSNPMissingRate=None, transferOutput=True,\
        maxContigID=None, outputDirPrefix="", outputPedigreeAsTFAM=False,
        outputPedigreeAsTFAMInputJobData=None, \
        treatEveryOneIndependent=False,\
        returnMode=3, ModifyTPEDRunType=1, chr_id2cumu_chr_start=None,\
        addUngenotypedDuoParents=False):
        """
    returnMode
        1: only the final merged binary .bed , .fam file and its generation job(s)
        2: only the individual contig/chromosome (whatever in inputDat.jobDataLs)
             binary .bed, .fam files and the associated jobs
        3: 1 & 2 (all individual binary .bed jobs&files + the last merged file/job)
        4: the individual contig/chr non-binary (TPED) job data
         (for Mark mendel error genotype as missing)
        5: 
        
        2013.07.18
            added argument addUngenotypedDuoParents
                for mendel error detection, if an ungenotyped parent in a duo
                 is not present in the genotype file (PED/TPED/BED),
                then plink won't look for its mendel inconsistency 
        2013.02
        2013.1.29 added returnMode 4
        2012.10.22
            change transferOutput of outputPedigreeInTFAMJob to True
        2012.9.13
            add argument treatEveryOneIndependent for
             OutputVRCPedigreeInTFAMGivenOrderFromFile.
        2012.8.20 add outputPedigreeAsTFAMInputJobData, split from input_data.
            outputPedigreeAsTFAMInputJobData.vcfFile must use
             individual_alignment.read_group as sample ID.
            useful in the case of that VCF files have been converted into UCLA IDs.
        2012.8.9
            add argument
                outputPedigreeAsTFAM
                returnMode
            add plink binary 
        2012.7.19 
            add a modifyTPEDJob that modify 2nd column (snp-id) of tped output from default 0 to chr_pos.
                argument ModifyTPEDRunType.
                1: modify snp_id (2nd-column) = chr_phyiscalPosition,\
                2: snp_id=chr_physicalPosition (original data), chr (1st column) = X (chromosome X, for sex check by plink), pos += positionStartBase.,\
                3: snp_id=chr_physicalPosition (original data), chr (1st column) = newChr, pos += positionStartBase
            
            added a GzipSubworkflow in the end to gzip the final merged tped file
            all previous intermediate files are not transferred.
        2012.5.9
        """
        sys.stderr.write("Adding VCF2plink jobs for %s vcf files ... "%(\
            len(inputData.jobDataLs)))
        
        topOutputDir = "%sVCF2Plink"%(outputDirPrefix)
        topOutputDirJob = self.addMkDirJob(outputDir=topOutputDir)
        
        mergedOutputDir = "%sVCF2PlinkMerged"%(outputDirPrefix)
        mergedOutputDirJob = self.addMkDirJob(outputDir=mergedOutputDir)
        
        mergedPlinkFnamePrefix = os.path.join(mergedOutputDir, 'merged')
        mergedTPEDFile = File('%s.tped'%(mergedPlinkFnamePrefix))
        #each input has no header
        tpedFileMergeJob = self.addStatMergeJob(
            statMergeProgram=self.mergeSameHeaderTablesIntoOne, \
            outputF=mergedTPEDFile, transferOutput=False,
            parentJobLs=[mergedOutputDirJob], \
            extraArguments='--noHeader')
        
        if outputPedigreeAsTFAMInputJobData is None:
            outputPedigreeAsTFAMInputJobData = inputData.jobDataLs[0]
        if outputPedigreeAsTFAM and outputPedigreeAsTFAMInputJobData:
            jobData = outputPedigreeAsTFAMInputJobData
            inputF = jobData.vcfFile
            outputFile = File(os.path.join(mergedOutputDir, 'pedigree.tfam'))
            outputPedigreeInTFAMJob = self.addOutputVRCPedigreeInTFAMGivenOrderFromFileJob(
                executable=self.OutputVRCPedigreeInTFAMGivenOrderFromFile, \
                inputFile=inputF, outputFile=outputFile,
                treatEveryOneIndependent=treatEveryOneIndependent,\
                addUngenotypedDuoParents=addUngenotypedDuoParents,\
                parentJobLs=[mergedOutputDirJob]+jobData.jobLs,
                extraDependentInputLs=[], transferOutput=True, \
                extraArguments=None, job_max_memory=2000,
                sshDBTunnel=self.needSSHDBTunnel)
            outputPedigreeInTFAMJob.tfamFile = outputPedigreeInTFAMJob.output
            #so that it looks like a vcf2plinkJob (vcftools job)
        else:
            outputPedigreeInTFAMJob = None
        
        returnData = PassingData()
        returnData.jobDataLs = []
        returnData.tfamJob = None
        #2013.07.25 family file for tped file 
        returnData.famJob = None
        #2013.07.25 family file for bed file
        
        for i in range(len(inputData.jobDataLs)):
            jobData = inputData.jobDataLs[i]
            inputF = jobData.vcfFile
            inputFBaseName = os.path.basename(inputF.name)
            chr_id = self.getChrFromFname(inputFBaseName)
            if maxContigID:
                contig_id = self.getContigIDFromFname(inputFBaseName)
                try:
                    contig_id = int(contig_id)
                    if contig_id>maxContigID:
                        #skip the small contigs
                        continue
                except:
                    sys.stderr.write('Except type: %s\n'%repr(sys.exc_info()))
                    import traceback
                    traceback.print_exc()
            commonPrefix = inputFBaseName.split('.')[0]
            outputFnamePrefix = os.path.join(topOutputDir, '%s'%(commonPrefix))
            if i ==0:
                #need at least one tfam file. 
                transferOneContigPlinkOutput = True
            else:
                transferOneContigPlinkOutput = False
            i += 1
            vcf2plinkJob = self.addFilterJobByvcftools(
                vcftoolsWrapper=self.vcftoolsWrapper,
                inputVCFF=inputF, \
                outputFnamePrefix=outputFnamePrefix, \
                parentJobLs=[topOutputDirJob]+jobData.jobLs, \
                snpMisMatchStatFile=None, \
                minMAC=minMAC, minMAF=minMAF, \
                maxSNPMissingRate=maxSNPMissingRate,\
                extraDependentInputLs=[jobData.tbi_F],
                outputFormat='--plink-tped',
                transferOutput=transferOneContigPlinkOutput)
            #2013.07.19
            if addUngenotypedDuoParents and outputPedigreeInTFAMJob:
                appendExtraIndividualsJob = self.addAbstractMapperLikeJob(
                    executable=self.AppendExtraPedigreeIndividualsToTPED, \
                    inputF=vcf2plinkJob.tpedFile, \
                    outputF=File(os.path.join(topOutputDir, '%s_extraIndividuals.tped'%(commonPrefix))),
                    parentJobLs=[vcf2plinkJob, outputPedigreeInTFAMJob, topOutputDirJob],
                    transferOutput=False, job_max_memory=200,
                    extraArgumentList=["--tfamFname", outputPedigreeInTFAMJob.tfamFile],
                    extraDependentInputLs=[outputPedigreeInTFAMJob.tfamFile])
                modifyTPEDParentJobLs = [appendExtraIndividualsJob]
                modifyTPEDJobInputFile = appendExtraIndividualsJob.output
            else:
                if addUngenotypedDuoParents and outputPedigreeInTFAMJob is None:
                    message = "Warning: addUngenotypedDuoParents (%s) is True but outputPedigreeInTFAMJob (%s, outputPedigreeAsTFAM=%s) is None. so addUngenotypedDuoParents is effectively False."%\
                            (addUngenotypedDuoParents, outputPedigreeInTFAMJob, outputPedigreeAsTFAM)
                    utils.pauseForUserInput(message=message, continueAnswerSet=None, exitType=3)
                modifyTPEDParentJobLs = [vcf2plinkJob]
                modifyTPEDJobInputFile = vcf2plinkJob.tpedFile
            
            #2012.7.20 modify the TPED 2nd column, to become chr_pos (rather than 0)
            modifyTPEDFnamePrefix = os.path.join(topOutputDir, '%s_SNPID_M'%(commonPrefix))
            outputF = File('%s.tped'%(modifyTPEDFnamePrefix))
            modifyTPEDJobExtraArguments = "--run_type %s "%(ModifyTPEDRunType)
            if ModifyTPEDRunType==3 and chr_id2cumu_chr_start:
                newChrID, newCumuStart = chr_id2cumu_chr_start.get(chr_id, (1,0))
                modifyTPEDJobExtraArguments += ' --newChr %s --positionStartBase %s '%(newChrID, newCumuStart)
            modifyTPEDJob = self.addAbstractMapperLikeJob(
                executable=self.ModifyTPED,
                inputF=modifyTPEDJobInputFile, outputF=outputF, \
                parentJobLs=modifyTPEDParentJobLs, transferOutput=False,
                job_max_memory=200,\
                extraArguments=modifyTPEDJobExtraArguments,
                extraDependentInputLs=[])
            
            #add output to the tped merge job
            self.addInputToMergeJob(tpedFileMergeJob, \
                inputF=modifyTPEDJob.output, \
                parentJobLs=[modifyTPEDJob])

            if outputPedigreeInTFAMJob is None:
                tfamJob = vcf2plinkJob
                convertSingleTPED2BEDParentJobLs = [modifyTPEDJob, vcf2plinkJob]
            else:
                tfamJob = outputPedigreeInTFAMJob
                convertSingleTPED2BEDParentJobLs = [modifyTPEDJob, outputPedigreeInTFAMJob]
            
            if returnData.tfamJob is None:
                returnData.tfamJob = tfamJob
            
            if returnMode==4:
                returnData.jobDataLs.append(PassingData(jobLs=[modifyTPEDJob], file=modifyTPEDJob.output, \
                    fileLs=modifyTPEDJob.outputLs))
            elif returnMode==2 or returnMode==3:
                #convert single plink tped file into binary bed file
                #add it to 
                bedFnamePrefix = os.path.join(topOutputDir, '%s_bed'%(commonPrefix))
                convertSingleTPED2BEDJob = self.addPlinkJob(
                    executable=self.plinkConvert, inputFileList=[], 
                    tpedFile=modifyTPEDJob.output, tfamFile=tfamJob.tfamFile,\
                    inputFnamePrefix=None, inputOption=None, \
                    outputFnamePrefix=bedFnamePrefix, outputOption='--out',\
                    makeBED=True, \
                    extraDependentInputLs=None, transferOutput=transferOutput, \
                    extraArguments=None, job_max_memory=2000,\
                    parentJobLs = convertSingleTPED2BEDParentJobLs)
                returnData.jobDataLs.append(PassingData(jobLs=[convertSingleTPED2BEDJob], file=convertSingleTPED2BEDJob.bedFile, \
                                            fileLs=convertSingleTPED2BEDJob.outputLs))
                if returnData.famJob is None:
                    returnData.famJob = convertSingleTPED2BEDJob
        
        if returnMode==1 or returnMode==3:
            #convert merged plain tped file into binary bed files
            mergedPlinkBEDFnamePrefix = os.path.join(mergedOutputDir, 'mergedPlinkBED')
            convertMergedTPED2BEDJob = self.addPlinkJob(
                executable=self.plinkMerge, inputFileList=[], \
                tpedFile=tpedFileMergeJob.output, tfamFile=tfamJob.tfamFile,\
                inputFnamePrefix=None, inputOption=None, \
                outputFnamePrefix=mergedPlinkBEDFnamePrefix,
                outputOption='--out',\
                makeBED=True, \
                extraDependentInputLs=None, transferOutput=transferOutput,
                extraArguments=None, job_max_memory=2000,
                parentJobLs=[mergedOutputDirJob, tpedFileMergeJob, tfamJob])
            returnData.jobDataLs.append(PassingData(jobLs=[convertMergedTPED2BEDJob],
                file=convertMergedTPED2BEDJob.bedFile,
                fileLs=convertMergedTPED2BEDJob.outputLs))
            if returnData.famJob is None:
                returnData.famJob = convertMergedTPED2BEDJob
        ##2012.8.9 gzip workflow is not needed anymore as binary bed is used instead.
        ##2012.7.21 gzip the final output
        gzipInputData = PassingData()
        gzipInputData.jobDataLs = []
        gzipInputData.jobDataLs.append(PassingData(jobLs=[tpedFileMergeJob],
            file=tpedFileMergeJob.output, \
            fileLs=tpedFileMergeJob.outputLs))
        self.addGzipSubWorkflow(inputData=gzipInputData, transferOutput=transferOutput,\
            outputDirPrefix="gzipMergedTPED")
        sys.stderr.write("%s jobs.\n"%(self.no_of_jobs))
        #2013.1.29 return the un-gzipped data so that downstream sub-workflows could work on un-gzipped files
        return returnData
    
    def addVCF2YuFormatJobs(self, inputData=None, transferOutput=True,\
                        maxContigID=None, outputDirPrefix="", \
                        returnMode=1):
        """
        2012.8.20
            argument
                returnMode
                    1=only the final merged file and its generation job(s)
                    2=only the individual contig/chromosome (whatever in inputDat.jobDataLs) converted files and conversion jobs
                    3= 1 & 2 (all individual input binary .bed job&file + the last merging job/file)
        """
        sys.stderr.write("Adding VCF2YuFormat jobs for %s vcf files ... "%(len(inputData.jobDataLs)))
        
        topOutputDir = "%sVCF2BjarniFormat"%(outputDirPrefix)
        topOutputDirJob = self.addMkDirJob(outputDir=topOutputDir)
        
        mergeOutputDir = "%sVCF2YuFormat"%(outputDirPrefix)
        mergeOutputDirJob = self.addMkDirJob(outputDir=mergeOutputDir)
        
        mergeFnamePrefix = os.path.join(mergeOutputDir, 'merged')
        mergeFile = File('%s.csv'%(mergeFnamePrefix))
        #each input has no header
        mergeFileJob = self.addStatMergeJob(statMergeProgram=self.mergeSameHeaderTablesIntoOne, \
                            outputF=mergeFile, transferOutput=False, parentJobLs=[mergeOutputDirJob])
        
        returnData = PassingData()
        returnData.jobDataLs = []
        for i in range(len(inputData.jobDataLs)):
            jobData = inputData.jobDataLs[i]
            inputF = jobData.vcfFile
            inputFBaseName = os.path.basename(inputF.name)
            chr_id = self.getChrFromFname(inputFBaseName)
            if maxContigID:
                contig_id = self.getContigIDFromFname(inputFBaseName)
                try:
                    contig_id = int(contig_id)
                    if contig_id>maxContigID:
                        #skip the small contigs
                        continue
                except:
                    sys.stderr.write('Except type: %s\n'%repr(sys.exc_info()))
                    import traceback
                    traceback.print_exc()
            commonPrefix = inputFBaseName.split('.')[0]
            outputFnamePrefix = os.path.join(topOutputDir, '%s'%(commonPrefix))
            if i ==0:
                #need at least one tfam file. 
                transferOneContigPlinkOutput = True
            else:
                transferOneContigPlinkOutput = False
            i += 1
            bjarniFormatOutput = File('%s.csv'%(outputFnamePrefix))
            vcf2BjarniFormatJob = self.addGenericJob(
                executable=self.ConvertVCF2BjarniFormat,
                inputFile=inputF, inputArgumentOption="-i", \
                outputFile=bjarniFormatOutput, outputArgumentOption="-o", \
                parentJobLs=[topOutputDirJob] + jobData.jobLs,
                extraDependentInputLs=None, extraOutputLs=None, \
                transferOutput=transferOneContigPlinkOutput, \
                extraArguments="--outputDelimiter ,", extraArgumentList=None,
                job_max_memory=2000, sshDBTunnel=None, \
                key2ObjectForJob=None)
            
            #add output to some reduce job
            self.addInputToMergeJob(mergeFileJob, \
                inputF=vcf2BjarniFormatJob.output, \
                parentJobLs=[vcf2BjarniFormatJob])
            
            
            if returnMode==2 or returnMode==3:
                returnData.jobDataLs.append(PassingData(jobLs=[vcf2BjarniFormatJob],
                    file=vcf2BjarniFormatJob.output, \
                    fileLs=vcf2BjarniFormatJob.outputLs))
            

        if returnMode==1 or returnMode==3:
            #conver the merged bjarni format into yu format
#			outputFile = File(os.path.join(mergeOutputDir, 'merge.tsv'))
#			bjar2YuFormatJob = self.addGenericJob(
#               executable=self.ConvertBjarniSNPFormat2Yu,
#               inputFile=mergeFileJob.output, inputArgumentOption="-i", \
#				outputFile=outputFile, outputArgumentOption="-o", \
#				parentJobLs=[mergeFileJob], extraDependentInputLs=None, extraOutputLs=None, \
#				transferOutput=transferOutput, \
#				extraArguments=None, extraArgumentList=None, job_max_memory=8000,
#               sshDBTunnel=None, \
#				key2ObjectForJob=None)
#			no_of_jobs += 1
#			returnData.jobDataLs.append(PassingData(jobLs=[bjar2YuFormatJob],
#               file=bjar2YuFormatJob.output, \
#				fileLs=bjar2YuFormatJob.outputLs))
            pass	#too much memory
        sys.stderr.write("%s jobs. Done.\n"%(self.no_of_jobs))
        return returnData
    
    def addVCFSubsetJob(self, executable=None, vcfSubsetPath=None, 
        sampleIDFile=None,
        inputVCF=None, outputF=None, \
        parentJobLs=None, transferOutput=True, job_max_memory=200,\
        extraArguments=None, extraDependentInputLs=None, **keywords):
        """
        2012.10.10 use addGenericJob
        """
        extraArgumentList = [vcfSubsetPath, sampleIDFile, inputVCF, outputF]
        if extraDependentInputLs is None:
            extraDependentInputLs = []
        extraDependentInputLs.append(sampleIDFile)
        extraDependentInputLs.append(inputVCF)
        extraOutputLs = [outputF]
        job = self.addGenericJob(executable=executable, inputFile=None, \
            outputFile=None, \
            parentJobLs=parentJobLs,
            extraDependentInputLs=extraDependentInputLs,
            extraOutputLs=extraOutputLs, \
            transferOutput=transferOutput, \
            extraArguments=extraArguments, extraArgumentList=extraArgumentList,
            job_max_memory=job_max_memory, sshDBTunnel=None, \
            key2ObjectForJob=None, **keywords)
        return job

    def addVCFSubsetJobs(self, inputData=None, db_main=None,
        sampleIDFile=None, transferOutput=True,\
        refFastaFList=None, GenomeAnalysisTKJar=None,\
        maxContigID=None, outputDirPrefix=""):
        """
        2012.10.5
            add a GATK SelectVariants job to update AC/AF of the final VCF file
            add argument refFastaFList, GenomeAnalysisTKJar
            
        2012.5.9
        """
        if GenomeAnalysisTKJar is None:
            GenomeAnalysisTKJar = self.GenomeAnalysisTKJar
        if refFastaFList is None:
            refFastaFList = self.refFastaFList
        
        sys.stderr.write("Adding vcf-subset jobs for %s vcf files ... "%(len(inputData.jobDataLs)))
        no_of_jobs= 0
        
        
        topOutputDir = "%sVCFSubset"%(outputDirPrefix)
        topOutputDirJob = self.addMkDirJob(outputDir=topOutputDir)
        no_of_jobs += 1
        
        returnData = PassingData()
        returnData.jobDataLs = []
        for jobData in inputData.jobDataLs:
            inputF = jobData.vcfFile
            chr = self.getChrFromFname(inputF.name)
            if maxContigID:
                contig_id = self.getContigIDFromFname(inputF.name)
                try:
                    contig_id = int(contig_id)
                    if contig_id>maxContigID:	#skip the small contigs
                        continue
                except:
                    sys.stderr.write('Except type: %s\n'%repr(sys.exc_info()))
                    import traceback
                    traceback.print_exc()
            inputFBaseName = os.path.basename(inputF.name)
            commonPrefix = inputFBaseName.split('.')[0]
            outputVCF = File(os.path.join(topOutputDir, '%s.subset.vcf'%(commonPrefix)))
            vcfSubsetJob = self.addVCFSubsetJob(executable=self.vcfSubset,
                vcfSubsetPath=self.vcfSubsetPath,
                sampleIDFile=sampleIDFile,\
                inputVCF=inputF, outputF=outputVCF, \
                parentJobLs=[topOutputDirJob]+jobData.jobLs,
                transferOutput=False, job_max_memory=200,\
                extraArguments=None, extraDependentInputLs=None)
            
            #2012.10.5
            #selectVariants would generate AC, AF so that TrioCaller could read it.
            #samtools uses 'AC1' instead of AC, 'AF1' instead of AF.
            VCF4OutputF = File(os.path.join(topOutputDir, '%s.niceformat.vcf'%commonPrefix))
            vcfConvertJob = self.addSelectVariantsJob(
                SelectVariantsJava=self.SelectVariantsJava, \
                inputF=vcfSubsetJob.output, outputF=VCF4OutputF, \
                refFastaFList=refFastaFList, parentJobLs=[vcfSubsetJob], \
                extraDependentInputLs=[], transferOutput=False, \
                extraArguments=None, job_max_memory=2000, interval=chr)
            
            VCFGzipOutputF = File("%s.gz"%VCF4OutputF.name)
            VCFGzipOutput_tbi_F = File("%s.gz.tbi"%VCF4OutputF.name)
            bgzip_tabix_VCF_job = self.addBGZIP_tabix_Job(
                bgzip_tabix=self.bgzip_tabix, \
                parentJobLs=[vcfConvertJob], inputF=vcfConvertJob.output,
                outputF=VCFGzipOutputF, \
                transferOutput=transferOutput)
            
            returnData.jobDataLs.append(
                PassingData(jobLs=[bgzip_tabix_VCF_job], vcfFile=VCFGzipOutputF, \
                    tbi_F=VCFGzipOutput_tbi_F, \
                    fileLs=[VCFGzipOutputF, VCFGzipOutput_tbi_F]))
            
        sys.stderr.write("%s jobs.\n"%(self.no_of_jobs))
        return returnData
    

    def addSubsetAndVCF2PlinkJobs(self, inputData=None, db_main=None,
        minMAC=None, minMAF=None,\
        maxSNPMissingRate=None, sampleIDFile=None, transferOutput=True,\
        maxContigID=None, outputDirPrefix=""):
        """
        2012.5.9
        """
        vcfSubsetJobData = self.addVCFSubsetJobs(inputData=inputData, db_main=db_main,
            sampleIDFile=sampleIDFile, \
            transferOutput=True, maxContigID=maxContigID, outputDirPrefix="")
        vcf2plinkJobData = self.addVCF2PlinkJobs(inputData=vcfSubsetJobData,
            db_main=db_main, \
            minMAC=minMAC, minMAF=minMAF, maxSNPMissingRate=maxSNPMissingRate,
            transferOutput=transferOutput,\
            outputPedigreeAsTFAM=False, outputPedigreeAsTFAMInputJobData=None,
            maxContigID=maxContigID, outputDirPrefix="")
    
    def addAlignmentReadGroup2UCLAIDJobs(self, inputData=None, db_main=None,
        transferOutput=True,\
        maxContigID=None, outputDirPrefix=""):
        """
        2012.5.9
        """
        sys.stderr.write("Adding alignment read-group -> UCLAID jobs for %s vcf files ... "%(len(inputData.jobDataLs)))
        no_of_jobs= 0
        
        
        topOutputDir = "%sSampleInUCLAID"%(outputDirPrefix)
        topOutputDirJob = self.addMkDirJob(outputDir=topOutputDir)
        no_of_jobs += 1
        
        returnData = PassingData()
        returnData.jobDataLs = []
        for jobData in inputData.jobDataLs:
            inputF = jobData.vcfFile
            if maxContigID:
                contig_id = self.getContigIDFromFname(inputF.name)
                try:
                    contig_id = int(contig_id)
                    if contig_id>maxContigID:	#skip the small contigs
                        continue
                except:
                    sys.stderr.write('Except type: %s\n'%repr(sys.exc_info()))
                    import traceback
                    traceback.print_exc()
            inputFBaseName = os.path.basename(inputF.name)
            commonPrefix = inputFBaseName.split('.')[0]
            outputVCF = File(os.path.join(topOutputDir, '%s.UCLAID.vcf'%(commonPrefix)))
            abstractMapperJob = self.addAbstractMapperLikeJob(
                executable=self.ConvertAlignmentReadGroup2UCLAIDInVCF, \
                inputVCF=inputF, outputF=outputVCF, \
                parentJobLs=[topOutputDirJob]+jobData.jobLs,
                transferOutput=False, job_max_memory=1000,\
                extraArguments=None, extraDependentInputLs=[])
            
            VCFGzipOutputF = File("%s.gz"%outputVCF.name)
            VCFGzipOutput_tbi_F = File("%s.gz.tbi"%outputVCF.name)
            bgzip_tabix_VCF_job = self.addBGZIP_tabix_Job(bgzip_tabix=self.bgzip_tabix, \
                parentJobLs=[abstractMapperJob], inputF=abstractMapperJob.output,
                outputF=VCFGzipOutputF, \
                transferOutput=transferOutput)
            
            returnData.jobDataLs.append(PassingData(jobLs=[bgzip_tabix_VCF_job],
                vcfFile=VCFGzipOutputF, \
                tbi_F=VCFGzipOutput_tbi_F, \
                fileLs=[VCFGzipOutputF, VCFGzipOutput_tbi_F]))
            
        sys.stderr.write("%s jobs.\n"%(self.no_of_jobs))
        return returnData
    
    def addSplitNamVCFJobs(self, inputData=None, db_main=None, transferOutput=True,\
        maxContigID=None, outputDirPrefix=""):
        """
        2012.5.11
            not functional. don't know what to do the fact that SplitNamVCFIntoMultipleSingleChrVCF outputs into a folder
                multiple VCF files (one per chromosome)
        """
        sys.stderr.write("Adding split Nam VCF-file jobs for %s vcf files ... "%(len(inputData.jobDataLs)))
        no_of_jobs= 0
        
        
        topOutputDir = "%sSampleInUCLAID"%(outputDirPrefix)
        topOutputDirJob = self.addMkDirJob(outputDir=topOutputDir)
        no_of_jobs += 1
        
        returnData = PassingData()
        returnData.jobDataLs = []
        for jobData in inputData.jobDataLs:
            inputF = jobData.vcfFile
            if maxContigID:
                contig_id = self.getContigIDFromFname(inputF.name)
                try:
                    contig_id = int(contig_id)
                    if contig_id>maxContigID:	#skip the small contigs
                        continue
                except:
                    sys.stderr.write('Except type: %s\n'%repr(sys.exc_info()))
                    import traceback
                    traceback.print_exc()
            inputFBaseName = os.path.basename(inputF.name)
            commonPrefix = inputFBaseName.split('.')[0]
            outputVCF = File(os.path.join(topOutputDir, '%s.vcf'%(commonPrefix)))
            abstractMapperJob = self.addAbstractMapperLikeJob(executable=self.SplitNamVCFIntoMultipleSingleChrVCF, \
                    inputVCF=inputF, outputF=outputVCF, \
                    parentJobLs=[topOutputDirJob]+jobData.jobLs, transferOutput=False, job_max_memory=200,\
                    extraArguments=None, extraDependentInputLs=[])
            
            VCFGzipOutputF = File("%s.gz"%outputVCF.name)
            VCFGzipOutput_tbi_F = File("%s.gz.tbi"%outputVCF.name)
            bgzip_tabix_VCF_job = self.addBGZIP_tabix_Job(bgzip_tabix=self.bgzip_tabix, \
                    parentJobLs=[abstractMapperJob], inputF=abstractMapperJob.output,
                    outputF=VCFGzipOutputF, \
                    transferOutput=transferOutput)
            
            returnData.jobDataLs.append(PassingData(jobLs=[bgzip_tabix_VCF_job],
                vcfFile=VCFGzipOutputF, \
                tbi_F=VCFGzipOutput_tbi_F, \
                fileLs=[VCFGzipOutputF, VCFGzipOutput_tbi_F]))
            
            no_of_jobs += 2
        sys.stderr.write("%s jobs.\n"%(self.no_of_jobs))
        return returnData
    
    def addMergeVCFReplicateHaplotypesJobs(self, inputData=None, db_main=None,
        transferOutput=True,\
        maxContigID=None, outputDirPrefix="",replicateIndividualTag='copy',
        refFastaFList=None ):
        """
        2012.7.25
            input vcf is output of TrioCaller with replicates.
            this workflow outputs extra debug statistics
                1. replicate haplotype distance to the consensus haplotype
                2. majority support for the consensus haplotype
        """
        sys.stderr.write("Adding MergeVCFReplicateHaplotype jobs for %s vcf files ... "%(len(inputData.jobDataLs)))
        no_of_jobs= 0
        
        
        topOutputDir = "%sMergeVCFReplicateHaplotypeStat"%(outputDirPrefix)
        topOutputDirJob = self.addMkDirJob(outputDir=topOutputDir)
        no_of_jobs += 1
        
        
        haplotypeDistanceMergeFile = File(os.path.join(topOutputDir, 'haplotypeDistanceMerge.tsv'))
        haplotypeDistanceMergeJob = self.addStatMergeJob(self.mergeSameHeaderTablesIntoOne, \
            outputF=haplotypeDistanceMergeFile, transferOutput=False,
            parentJobLs=[topOutputDirJob])
        majoritySupportMergeFile = File(os.path.join(topOutputDir, 'majoritySupportMerge.tsv'))
        majoritySupportMergeJob = self.addStatMergeJob(self.mergeSameHeaderTablesIntoOne, \
            outputF=majoritySupportMergeFile, transferOutput=False,
            parentJobLs=[topOutputDirJob])
        no_of_jobs += 2
        
        returnData = PassingData()
        returnData.jobDataLs = []
        for jobData in inputData.jobDataLs:
            inputF = jobData.vcfFile
            inputFBaseName = os.path.basename(inputF.name)
            commonPrefix = inputFBaseName.split('.')[0]
            outputVCF = File(os.path.join(topOutputDir, '%s.vcf'%(commonPrefix)))
            debugHaplotypeDistanceFile = File(os.path.join(topOutputDir, '%s.haplotypeDistance.tsv'%(commonPrefix)))
            debugMajoritySupportFile = File(os.path.join(topOutputDir, '%s.majoritySupport.tsv'%(commonPrefix)))
            fileSize = utils.getFileOrFolderSize(pegaflow.getAbsPathOutOfFile(inputF))
            memoryRequest = 45000
            memoryRequest = min(42000, max(4000, int(38000*(fileSize/950452059.0))) )
            #extrapolates (33,000Mb memory for a ungzipped VCF file with size=950,452,059)
            #upper bound is 42g. lower bound is 4g.
            #mergeReplicateOutputF = File(os.path.join(trioCallerOutputDirJob.folder, '%s.noReplicate.vcf'%vcfBaseFname))
            #noOfAlignments= len(alignmentDataLs)
            #entireLength = stopPos - startPos + 1	#could be very small for shorter reference contigs
            #memoryRequest = min(42000, max(4000, int(20000*(noOfAlignments/323.0)*(entireLength/2600000.0))) )
                #extrapolates (20000Mb memory for a 323-sample + 2.6Mbase reference length/26K loci)
                #upper bound is 42g. lower bound is 4g.
            mergeVCFReplicateColumnsJob = self.addMergeVCFReplicateGenotypeColumnsJob(
                executable=self.MergeVCFReplicateHaplotypesJava,\
                GenomeAnalysisTKJar=self.GenomeAnalysisTKJar, \
                inputF=inputF, outputF=outputVCF,
                replicateIndividualTag=replicateIndividualTag,
                refFastaFList=refFastaFList,
                debugHaplotypeDistanceFile=debugHaplotypeDistanceFile,
                debugMajoritySupportFile=debugMajoritySupportFile,
                parentJobLs=[topOutputDirJob]+jobData.jobLs,
                extraDependentInputLs=[], transferOutput=False,
                extraArguments=None, job_max_memory=memoryRequest)
            
            #add output to some reduce job
            self.addInputToMergeJob(haplotypeDistanceMergeJob,
                inputF=mergeVCFReplicateColumnsJob.outputLs[1],
                parentJobLs=[mergeVCFReplicateColumnsJob])
            self.addInputToMergeJob(majoritySupportMergeJob,
                inputF=mergeVCFReplicateColumnsJob.outputLs[2],
                parentJobLs=[mergeVCFReplicateColumnsJob])
            no_of_jobs += 1
        sys.stderr.write("%s jobs. Done.\n"%(no_of_jobs))
        
        returnData.jobDataLs.append(PassingData(jobLs=[haplotypeDistanceMergeJob],
            file=haplotypeDistanceMergeFile,
            fileLs=[haplotypeDistanceMergeFile]))
        returnData.jobDataLs.append(PassingData(jobLs=[majoritySupportMergeJob],
            file=majoritySupportMergeFile,
            fileLs=[majoritySupportMergeFile]))
        #2012.7.21 gzip the final output
        newReturnData = self.addGzipSubWorkflow(inputData=returnData, 
            transferOutput=transferOutput,\
            outputDirPrefix="")
        return newReturnData
    
    def generateVCFSampleIDFilenameFromIndividualUCLAIDFname(self,
        db_main=None, individualUCLAIDFname=None,
        vcfSampleIDFname=None, oneSampleVCFFname=None):
        """
            
        """
        sys.stderr.write("Generating vcfSampleIDFname %s from individualUCLAIDFname %s ..."%(\
            vcfSampleIDFname, individualUCLAIDFname))
        
        #first get the set of monkeys to keep from the file
        reader = csv.reader(open(individualUCLAIDFname), 
            delimiter=figureOutDelimiter(individualUCLAIDFname))
        header = next(reader)
        colName2Index = getColName2IndexFromHeader(header)
        UCLAID_col_index = colName2Index.get('UCLAID')
        individualUCLAIDSet = set()
        for row in reader:
            individualUCLAID=row[UCLAID_col_index].strip()
            individualUCLAIDSet.add(individualUCLAID)
        sys.stderr.write(" %s uclaIDs. "%(len(individualUCLAIDSet)))
        del reader
        
        #second, read a sample VCF file and output the samples that have 
        # been in the given set
        writer = csv.writer(open(vcfSampleIDFname, 'w'), delimiter='\t')
        vcfFile = VCFFile(inputFname=oneSampleVCFFname, minDepth=0)
        no_of_samples = 0
        for sample_id in vcfFile.sample_id_ls:
            individual_code = db_main.parseAlignmentReadGroup(sample_id).individual_code
            if individual_code in individualUCLAIDSet:
                no_of_samples += 1
                writer.writerow([sample_id])
        del writer, vcfFile
        sys.stderr.write("%s vcf samples selected.\n"%(no_of_samples))
    
    def addCombineVCFIntoOneJobs(self, inputData=None, data_dir=None,
        maxContigID=None, outputDirPrefix="", genotypeMethodShortName=None,
        needSSHDBTunnel=False,
        transferOutput=True):
        """
        2012.8.30
        """
        sys.stderr.write("Adding jobs to combine all VCF files into one VCF for %s vcf files ... "%(len(inputData.jobDataLs)))
        no_of_jobs= 0
        
        topOutputDir = "%sVCFIntoOne"%(outputDirPrefix)
        topOutputDirJob = self.addMkDirJob(outputDir=topOutputDir)
        no_of_jobs += 1
        
        
        
        #2011-9-22 union of all samtools intervals for one contig
        if genotypeMethodShortName:
            #if this is available, output goes to db-affiliated storage.
            transferUnionOutput = False
        else:
            transferUnionOutput = transferOutput
        unionOutputFname = os.path.join(topOutputDir, 
            '%sVCFIntoOne.vcf.gz'%(len(inputData.jobDataLs)))
        unionOutputF = File(unionOutputFname)
        unionJob = self.addVCFConcatJob(
            concatExecutable=self.concatSamtools,
            parentDirJob=topOutputDirJob, \
            outputF=unionOutputF, transferOutput=transferUnionOutput,
            vcf_job_max_memory=2000)
        
        no_of_jobs += 1
        
        returnData = PassingData()
        returnData.jobDataLs = []
        returnData.jobDataLs.append(PassingData(jobLs=[unionJob],
            file=unionJob.output,
            fileLs=unionJob.outputLs))
        
        for i in range(len(inputData.jobDataLs)):
            jobData = inputData.jobDataLs[i]
            inputF = jobData.vcfFile
            inputFBaseName = os.path.basename(inputF.name)
            chr_id = self.getChrFromFname(inputFBaseName)
            if maxContigID:
                contig_id = self.getContigIDFromFname(inputFBaseName)
                try:
                    contig_id = int(contig_id)
                    if contig_id>maxContigID:
                        #skip the small contigs
                        continue
                except:
                    sys.stderr.write('Except type: %s\n'%repr(sys.exc_info()))
                    import traceback
                    traceback.print_exc()
            commonPrefix = inputFBaseName.split('.')[0]
            outputFnamePrefix = os.path.join(topOutputDir, '%s'%(commonPrefix))
            if i ==0:	#need at least one tfam file. 
                transferOneContigPlinkOutput = True
            else:
                transferOneContigPlinkOutput = False
            i += 1
            
            #add this output to a union job
            self.addInputToMergeJob(unionJob, inputF=inputF,
                parentJobLs=jobData.jobLs, extraDependentInputLs=[jobData.tbi_F])
            
        if genotypeMethodShortName:
            logFile = File(os.path.join(topOutputDir, 'addVCF2DB.log'))
            addVCFJob = self.addAddVCFFile2DBJob(executable=self.AddVCFFile2DB,
                inputFile=unionJob.output, \
                genotypeMethodShortName=genotypeMethodShortName,
                logFile=logFile, format="VCF", data_dir=data_dir,
                checkEmptyVCFByReading=True, commit=True,
                parentJobLs=[unionJob], extraDependentInputLs=[unionJob.tbi_F],
                transferOutput=transferOutput,
                extraArguments=None, job_max_memory=1000, sshDBTunnel=needSSHDBTunnel)
            no_of_jobs += 1
        sys.stderr.write("%s jobs. Done.\n"%(no_of_jobs))
        ##2012.8.9 gzip workflow is not needed anymore as binary bed is used instead.
        ##2012.7.21 gzip the final output
        #newReturnData = self.addGzipSubWorkflow(inputData=returnData,
        #   transferOutput=transferOutput,
        #				outputDirPrefix="")
        return returnData
        
    
    def registerExecutables(self):
        """
        2011-11-28
        """
        ParentClass.registerExecutables(self)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath,
            "mapper/ConvertAlignmentReadGroup2UCLAIDInVCF.py"),
            name='ConvertAlignmentReadGroup2UCLAIDInVCF', clusterSizeMultiplier=1)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath,
            "mapper/SplitNamVCFIntoMultipleSingleChrVCF.py"),
            name='SplitNamVCFIntoMultipleSingleChrVCF', clusterSizeMultiplier=1)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath,
            "mapper/filter/ModifyTPED.py"),
            name='ModifyTPED', clusterSizeMultiplier=1)
        
        self.registerOneExecutable(path=os.path.join(self.pymodulePath,
            "mapper/extractor/vcfSubset.sh"),\
            name='vcfSubset', clusterSizeMultiplier=1)
        if self.vcfSubsetPath:
            self.vcfSubset.vcfSubsetPath = self.vcfSubsetPath
            #vcfSubsetPath is first argument to vcfSubset
            self.vcfSubsetExecutableFile = self.registerOneExecutableAsFile(
                path=self.vcfSubsetPath)
        self.registerOneExecutable(path=self.plinkPath,
            name='plink', clusterSizeMultiplier=1)
        self.registerOneExecutable(path=self.plinkPath,
            name='plinkNoClustering', clusterSizeMultiplier=0)
        self.registerOneExecutable(path=self.plinkPath,
            name='plinkMerge', clusterSizeMultiplier=0)
        self.registerOneExecutable(path=self.plinkPath,
            name='plinkIBD', clusterSizeMultiplier=0)
        self.registerOneExecutable(path=self.plinkPath,
            name='plinkConvert', clusterSizeMultiplier=1)
        self.registerOneExecutable(path=self.plinkPath,
            name='plinkLDPrune', clusterSizeMultiplier=1)
        self.registerOneExecutable(path=self.plinkPath, \
            name='plinkExtract', clusterSizeMultiplier=1)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath,
            'mapper/modifier/SplitPlinkLMendelFileSNPIDIntoChrPosition.py'),
            name='SplitPlinkLMendelFileSNPIDIntoChrPosition', clusterSizeMultiplier=1)
        self.registerOneExecutable(path=os.path.join(self.pymodulePath,
            'pedigree/CalculateMendelErrorRateGivenPlinkOutput.py'),
            name='CalculateMendelErrorRateGivenPlinkOutput', clusterSizeMultiplier=1)

    
    def run(self):
        """
        2011-9-28
        """
        self.setup_run()
        inputData = self.registerAllInputFiles(self.inputDir, 
            input_site_handler=self.input_site_handler, \
            checkEmptyVCFByReading=self.checkEmptyVCFByReading,\
            pegasusFolderName=self.pegasusFolderName,\
            maxContigID=self.maxContigID, \
            minContigID=self.minContigID)
        if len(inputData.jobDataLs)<=0:
            sys.stderr.write("No VCF files in this folder , %s.\n"%self.inputDir)
            sys.exit(0)
        
        if self.individualUCLAIDFname and os.path.isfile(self.individualUCLAIDFname):
            self.generateVCFSampleIDFilenameFromIndividualUCLAIDFname(
                db_main=self.db_main,
                individualUCLAIDFname=self.individualUCLAIDFname,
                vcfSampleIDFname=self.vcfSampleIDFname,
                oneSampleVCFFname=inputData.jobDataLs[0].vcfFile.abspath)
            sampleIDFile = self.registerOneInputFile(self.vcfSampleIDFname)
        elif self.vcfSampleIDFname and os.path.isfile(self.vcfSampleIDFname):
            sampleIDFile = self.registerOneInputFile(self.vcfSampleIDFname)
        else:
            sampleIDFile = None
        
        if self.run_type==1:
            if sampleIDFile is None:
                sys.stderr.write("sampleIDFile is None.\n")
                sys.exit(0)
            self.addVCFSubsetJobs(
                inputData=inputData, db_main=self.db_main, sampleIDFile=sampleIDFile,
                transferOutput=True,\
                refFastaFList=self.refFastaFList,
                GenomeAnalysisTKJar=self.GenomeAnalysisTKJar,\
                maxContigID=self.maxContigID, outputDirPrefix="")
        elif self.run_type==2:
            #2012.8.10 test  ModifyTPEDRunType=3, chr_id2cumu_chr_start=None
            self.addVCF2PlinkJobs(
                inputData=inputData, db_main=self.db_main,
                minMAC=self.minMAC, minMAF=self.minMAF,\
                maxSNPMissingRate=self.maxSNPMissingRate, transferOutput=True,\
                outputPedigreeAsTFAM=False, outputPedigreeAsTFAMInputJobData=None, \
                maxContigID=self.maxContigID, outputDirPrefix="")
        elif self.run_type==3:
            if sampleIDFile is None:
                sys.stderr.write("sampleIDFile is None.\n")
                sys.exit(0)
            self.addSubsetAndVCF2PlinkJobs(
                inputData=inputData, db_main=self.db_main, minMAC=self.minMAC, \
                minMAF=self.minMAF,\
                maxSNPMissingRate=self.maxSNPMissingRate, sampleIDFile=sampleIDFile, transferOutput=True,\
                maxContigID=self.maxContigID, outputDirPrefix="")
        elif self.run_type==4:
            self.addAlignmentReadGroup2UCLAIDJobs(
                inputData=inputData, 
                db_main=self.db_main, transferOutput=True,\
                maxContigID=self.maxContigID, outputDirPrefix="")
        elif self.run_type==5:
            refSequence = SunsetDB.IndividualSequence.get(self.ref_ind_seq_id)
            refFastaFname = os.path.join(self.data_dir, refSequence.path)
            registerReferenceData = self.registerRefFastaFile(refFastaFname,
                registerAffiliateFiles=True,
                checkAffiliateFileExistence=True)
            self.addMergeVCFReplicateHaplotypesJobs(inputData=inputData, 
                db_main=self.db_main, transferOutput=True,\
                maxContigID=self.maxContigID, outputDirPrefix="", 
                replicateIndividualTag='copy', \
                refFastaFList=registerReferenceData.refFastaFList )
        elif self.run_type==6:
            self.addVCF2YuFormatJobs(inputData=inputData, transferOutput=True,\
                maxContigID=self.maxContigID, outputDirPrefix="", \
                returnMode=1)
        elif self.run_type==7:
            #first convert every sample ID from alignment.read_group to simple ucla ID 
            inputData2 = self.addAlignmentReadGroup2UCLAIDJobs(
                inputData=inputData, 
                db_main=self.db_main, \
                transferOutput=False,\
                maxContigID=self.maxContigID, outputDirPrefix="")
            self.addVCF2YuFormatJobs(inputData=inputData2, transferOutput=True,\
                maxContigID=self.maxContigID, outputDirPrefix="", \
                returnMode=1)
        elif self.run_type==8:
            inputData2 = self.addAlignmentReadGroup2UCLAIDJobs(inputData=inputData, 
                db_main=self.db_main, transferOutput=False,\
                maxContigID=self.maxContigID, outputDirPrefix="")
            self.addVCF2PlinkJobs(inputData=inputData2, db_main=self.db_main, \
                minMAC=self.minMAC, minMAF=self.minMAF,\
                maxSNPMissingRate=self.maxSNPMissingRate, transferOutput=True,\
                outputPedigreeAsTFAM=True,
                outputPedigreeAsTFAMInputJobData=inputData.jobDataLs[0],
                maxContigID=self.maxContigID, outputDirPrefix="", returnMode=3)
        elif self.run_type==9:
            self.addCombineVCFIntoOneJobs(inputData=inputData, data_dir=self.data_dir,
                maxContigID=self.maxContigID, outputDirPrefix="", 
                genotypeMethodShortName=self.genotypeMethodShortName, \
                needSSHDBTunnel=self.needSSHDBTunnel, \
                transferOutput=True)
        else:
            sys.stderr.write("run_type %s not supported.\n"%(self.run_type))
            sys.exit(0)
        
        self.end_run()
        


if __name__ == '__main__':
    from argparse import ArgumentParser
    ap = ArgumentParser()
    ap.add_argument("--plinkPath", type=str, default='bin/plink',
        help="Path to plink, "
        "http://pngu.mgh.harvard.edu/~purcell/plink/index.shtml "
        "If relative path, home folder is inserted in the front.")
    



"""
Some simple functions to generate colours.
"""
from matplotlib.numerix import asarray, asum
from matplotlib.mlab import linspace
from matplotlib.colors import colorConverter
from matplotlib.numerix import sum

def pastel(colour, weight=2.4):
    """ Convert colour into a nice pastel shade"""
    rgb = asarray(colorConverter.to_rgb(colour))
    # scale colour 
    maxc = max(rgb)
    if maxc < 1.0 and maxc > 0:
        # scale colour
        scale = 1.0 / maxc
        rgb = rgb * scale
    # now decrease saturation
    total = asum(rgb)
    slack = 0
    for x in rgb:
        slack += 1.0 - x

    # want to increase weight from total to weight
    # pick x s.t.  slack * x == weight - total
    # x = (weight - total) / slack
    x = (weight - total) / slack

    rgb = [c + (x * (1.0-c)) for c in rgb]

    return rgb

def get_colours(n):
    """ Return n pastel colours. """
    base = asarray([[1,0,0], [0,1,0], [0,0,1]])

    if n <= 3:
        return base[0:n]

    # how many new colours to we need to insert between
    # red and green and between green and blue?
    needed = (((n - 3) + 1) / 2, (n - 3) / 2)

    colours = []
    for start in (0, 1):
        for x in linspace(0, 1, needed[start]+2):
            colours.append((base[start] * (1.0 - x)) +
                           (base[start+1] * x))

    return [pastel(c) for c in colours[0:n]]




#!/usr/bin/env python3

from enthought.mayavi.tools.helper_functions import Pipeline, document_pipeline, bar_mode_dict
from enthought.mayavi.tools import tools
from enthought.traits.api import Array, Callable, CFloat, HasTraits, \
	List, Trait, Any, Instance, TraitError

from enthought.mayavi.tools.sources import vector_scatter, vector_field, scalar_scatter, \
			scalar_field, line_source, array2d_source, grid_source, \
			triangular_mesh_source, vertical_vectors_source

from enthought.mayavi.tools.modules import VectorsFactory, StreamlineFactory, GlyphFactory, \
	IsoSurfaceFactory, SurfaceFactory, ContourSurfaceFactory, \
	ImageActorFactory, glyph_mode_dict

class CustomBarChart(Pipeline):
	"""
	2012.2.21
		custom version of BarChart. It has two more keyword arguments, x_scale, y_scale,
			which is in charge of the lateral_scale in the X and Y direction.
		
	Plots vertical glyphs (like bars) scaled vertical, to do
	histogram-like plots.

	This functions accepts a wide variety of inputs, with positions given
	in 2-D or in 3-D.

	**Function signatures**::

		barchart(s, ...)
		barchart(x, y, s, ...)
		barchart(x, y, f, ...)
		barchart(x, y, z, s, ...)
		barchart(x, y, z, f, ...)

	If only one positional argument is passed, it can be a 1-D, 2-D, or 3-D
	array giving the length of the vectors. The positions of the data
	points are deducted from the indices of array, and an
	uniformly-spaced data set is created.

	If 3 positional arguments (x, y, s) are passed the last one must be
	an array s, or a callable, f, that returns an array. x and y give the
	2D coordinates of positions corresponding to the s values. 
	
	If 4 positional arguments (x, y, z, s) are passed, the 3 first are
	arrays giving the 3D coordinates of the data points, and the last one
	is an array s, or a callable, f, that returns an array giving the
	data value.
	"""

	_source_function = Callable(vertical_vectors_source)

	_pipeline = [VectorsFactory, ]

	mode = Trait('cube', bar_mode_dict,
					desc='The glyph used to represent the bars.')

	lateral_scale = CFloat(0.9, desc='The lateral scale of the glyph, '
				'in units of the distance between nearest points')

	def __call__(self, *args, **kwargs):
		""" Override the call to be able to scale automaticaly the axis.
		"""
		g = Pipeline.__call__(self, *args, **kwargs)
		gs = g.glyph.glyph_source
		# Use a cube source for glyphs.
		if not 'mode' in kwargs:
			gs.glyph_source = gs.glyph_list[-1]
		# Position the glyph tail on the point.
		gs.glyph_position = 'tail'
		gs.glyph_source.center = (0.0, 0.0, 0.5)
		g.glyph.glyph.orient = False
		if not 'color' in kwargs:
			g.glyph.color_mode = 'color_by_scalar'
		if not 'scale_mode' in kwargs:
			g.glyph.scale_mode = 'scale_by_vector_components'
		g.glyph.glyph.clamping = False
		x, y, z = g.mlab_source.x, g.mlab_source.y, g.mlab_source.z
		scale_factor = g.glyph.glyph.scale_factor* \
					tools._min_axis_distance(x, y, z)
		x_scale = kwargs.pop('x_scale', self.lateral_scale)
		y_scale = kwargs.pop('y_scale', self.lateral_scale)
		try:
			g.glyph.glyph_source.glyph_source.y_length = \
					y_scale/(scale_factor)
			g.glyph.glyph_source.glyph_source.x_length = \
					x_scale/(scale_factor)
		except TraitError:
			" Not all types of glyphs have controlable y_length and x_length"

		return g
	
	def get_all_traits(self):
		"""
		2012.2.21
			this function determines the set of possible keys in kwargs.
			add two keywords, x_scale, y_scale.
			
		Returns all the traits of class, and the classes in the pipeline.
		"""
		#call the parental one first.
		traits = Pipeline.get_all_traits(self)
		
		traits['x_scale'] = Trait(0.9, desc='The scale of the glyph on the x-axis, '
				'in units of the distance between nearest points')
		traits['y_scale'] = Trait(0.9, desc='The scale of the glyph on the y-axis, '
				'in units of the distance between nearest points')
		return traits

customBarchart = document_pipeline(CustomBarChart())

def testVerticalSurface():
	import pylab, numpy
	pylab.clf()
	
	no_of_phenotypes = 5
	x, y = numpy.mgrid[0:2*no_of_phenotypes:1, 0:10:1]	#added a gap of 1 column between two phenotypes. one phenotype occupies two rows & two columns.
	
	
	#remove the gap in x & y
	needed_index_ls = [0,5]
	#for i in [0, 5]:
	#	for j in range(no_of_phenotypes):
	#		needed_index_ls.append(no_of_phenotypes*i+j)
	#for i in range(0, no_of_phenotypes):
	#	needed_index_ls.append(2*i)
		#needed_index_ls.append(3*i+1)
		#y[3*i+1][1]=2
	x = x[:, needed_index_ls]
	y = y[:, needed_index_ls]
	
	enrichment_matrix = numpy.ones(x.shape, numpy.float)
	enrichment_matrix[:,:] =10
	enrichment_matrix[0,0]=3
	
	from enthought.mayavi import mlab
	mlab.clf()
	#from palos.yh_mayavi import customBarchart
	bar = customBarchart(x, y , enrichment_matrix, x_scale=0.9, y_scale=4.5, opacity=1, mode='cube', color=(0,1,0), scale_factor=1.0)
	"""
	#mlab.ylabel("KW")
	#mlab.xlabel("Emma")
	#mlab.zlabel("Enrichment Ratio")
	from palos.DrawMatrix import get_font 
	font = get_font()
	
	for i in range(len(xlabel_ls)):
		label = xlabel_ls[i]
		char_width, char_height = font.getsize(label)	#W is the the biggest(widest)
		
		mlab.text(2*i, 0, label, z=0, width=char_width/1500.)	#min(0.0075*len(label), 0.04))
	
	"""
	s = numpy.zeros(x.shape, numpy.int)
	#s[0,1]=0.5
	surf = mlab.surf(x, y, s, opacity=0.6, extent=[-1, 2*no_of_phenotypes, -1, 10, 0.0,0.0])
	mlab.show()
	
if __name__ == '__main__':
	testVerticalSurface()